#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   clustering.py
@Time    :   2023/06/12 14:10:48
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   Relevant clustering functions for Network project. The functions were taken from the BU/Subtyping project.
"""
import warnings
from collections import Counter

import numpy as np
import pandas as pd
import plotly.express as px

# plotting
import plotly.graph_objects as go

# from fcmeans import FCM
from plotly.subplots import make_subplots
from sklearn import cluster, metrics, mixture
from sklearn.decomposition import PCA
from sklearn.neighbors import kneighbors_graph

colours_pool = (
    px.colors.qualitative.Bold
    + px.colors.qualitative.D3
    + px.colors.qualitative.G10
    + px.colors.qualitative.T10
    + px.colors.qualitative.Pastel
    + px.colors.qualitative.Prism
    + px.colors.qualitative.Vivid
    + px.colors.qualitative.Alphabet
)


# Functions from utilities on clustering


def clustering_methods(datasets, default_base, samples, selected_clusters=None):
    """This function is the core of applying different clustering methods to a dataset. It can run different experiments with different datasets and configuration for the algorithms. It's a modification of the scikit-learn [blogpost](https://scikit-learn.org/stable/modules/clustering.html). The following clusters are supported (name, id):

        1. Kmeans - RawKMeans"
        2. Mini Batch KMeans - MiniBatchKMeans
        3. Ward - Ward
        4. Birch - Birch
        5. Gaussian Mixture Models - GaussianMixture
        6. Affinity Propagation - AffinityPropagation
        7. SpectralClustering - Spectral Clustering
        8. DBSCAN - DBSCAN
        9. OPTICS - OPTICS
        10. Hierarchical Clustering - Hierarchical Clustering

    Args:
        datasets ([dic]): List of the touples which containes the datasets, which is a DataFrame and the cluster models parameters in the form of a dictionary that needs to be override. See default_base of the parameters that can be override.
        default_base ([dict]): The configurations to be override for an experiments. Defaults to 3 clusters and birch_th 1.7. In case it needs to be override, below are the acceptable parameters and defualt values:
                    {'quantile': .2,
                    'eps': .3,
                    'damping': .5,
                    'preference': 5,
                    'n_neighbors': 5,
                    'n_clusters': 3,
                    'min_samples': 5,
                    'xi': 0.05,
                    'min_cluster_size': 0.1,
                    "birch_th": 1.7,
                    'name': "Name of the experiment" }
        samples ([String]): The list of samples
        selected_clusters ([String], optional): List of the strings that are the cluster models supported. Defaults to None, which means that all the available cluster models will be used.

    Returns:
        [list]: List of touples which contains the following data (in  this order): the experiment name, cluster model output and the cluster model object itself.
    """

    # for backwards compatibility
    if "fuzziness" != default_base:
        default_base["fuzziness"] = 1.15

    # array which contains a list with the name of dataset, the cluster models and the output labels
    ret_val = []
    for dataset, algo_params in datasets:
        # update parameters with dataset-specific values
        params = default_base.copy()
        params.update(algo_params)

        X = dataset.values
        # connectivity matrix for structured Ward
        connectivity = kneighbors_graph(X, n_neighbors=params["n_neighbors"], include_self=False)
        # make connectivity symmetric
        connectivity = 0.5 * (connectivity + connectivity.T)

        # ============
        # Create cluster objects
        # ============
        two_means = cluster.MiniBatchKMeans(n_clusters=params["n_clusters"])
        kmeans = cluster.KMeans(n_clusters=params["n_clusters"], max_iter=1000, random_state=10, n_init=10)
        ward = cluster.AgglomerativeClustering(n_clusters=params["n_clusters"], linkage="ward", connectivity=connectivity)
        birch = cluster.Birch(n_clusters=params["n_clusters"], threshold=params["birch_th"])
        affinity_propagation = cluster.AffinityPropagation(damping=params["damping"], preference=params["preference"])
        # we know from prev experiments this is a good configuration
        gmm = mixture.GaussianMixture(n_components=params["n_clusters"], covariance_type="diag", max_iter=500)
        spectral = cluster.SpectralClustering(n_clusters=params["n_clusters"], eigen_solver="arpack", affinity="nearest_neighbors")
        dbscan = cluster.DBSCAN(eps=params["eps"], min_samples=params["min_samples"])
        optics = cluster.OPTICS(min_samples=params["min_samples"], xi=params["xi"], min_cluster_size=params["min_cluster_size"])
        hierarchical_clutering = cluster.AgglomerativeClustering(distance_threshold=0, n_clusters=None, linkage="average", connectivity=connectivity)

        average_linkage = cluster.AgglomerativeClustering(linkage="average", metric=params["affinity"], n_clusters=params["n_clusters"])

        # fcm = FCM(n_clusters=params["n_clusters"], m=params["fuzziness"])

        # average_linake_2 = cluster.AgglomerativeClustering(linkage="average", n_clusters=params["n_clusters"]);

        clustering_algorithms = (
            ("RawKMeans", kmeans),
            # ("FuzzyCMeans", fcm),
            ("MiniBatchKMeans", two_means),
            ("Ward", ward),
            ("Birch", birch),
            ("GaussianMixture", gmm),
            ("AffinityPropagation", affinity_propagation),
            ("SpectralClustering", spectral),
            ("Avg", average_linkage),
            # ('AggCluster', average_linkage),
            ("DBSCAN", dbscan),
            ("OPTICS", optics),
            ("Hierarchical Clustering", hierarchical_clutering),
        )

        # run only the ones we're interested
        clustering_algorithms = [clustering_algorithm for clustering_algorithm in clustering_algorithms if clustering_algorithm[0] in selected_clusters]

        output_algorithms = pd.DataFrame(samples, columns=["samples"])
        output_models = []

        for name, algorithm in clustering_algorithms:
            # selected_clusters is None means that all are selected
            if selected_clusters is not None:
                if name not in selected_clusters:
                    continue  # skip if it's not in the loop
            # catch warnings related to kneighbors_graph
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message="the number of connected components of the "
                    + "connectivity matrix is [0-9]{1,2}"
                    + " > 1. Completing it to avoid stopping the tree early.",
                    category=UserWarning,
                )
                warnings.filterwarnings(
                    "ignore", message="Graph is not fully connected, spectral embedding" + " may not work as expected.", category=UserWarning
                )
                algorithm.fit(X)

            if hasattr(algorithm, "labels_"):
                y_pred = algorithm.labels_.astype(int)
            else:
                y_pred = algorithm.predict(X)

            output_algorithms[name] = y_pred
            output_models.append((name, algorithm))

        # this need refactoring to returning a dictionary with key, pairs as it will be clearer
        ret_val.append((params["name"], output_algorithms, output_models))
    return ret_val


def compare_exp(
    data, rob_comp=None, n_clusters=None, default_base=None, selected_clusters=None, show_figures=True, show_consensus=True, pca_data=False, n_comp=2
):
    data = data.sort_index()
    # Process results
    samples = list(data.index.values)
    # sel_cols = data.columns.values

    # Show output
    pca_df, pca_plot = apply_pca(df=data, pca_components=2, transpose=True, samples_col="Sample", genes_col="Genes")

    pca_model = {}
    if pca_data:
        pca = PCA(n_components=n_comp)
        initial_data = data
        data = pd.DataFrame(pca.fit_transform(data), index=initial_data.index)
        pca_model["pca"] = pca
        pca_model["data"] = data
        print("PC var {} with a total of {:.02f}%".format(pca.explained_variance_ratio_, np.sum(pca.explained_variance_ratio_) * 100))
        print("PCA score ", pca.score(initial_data.values))

    datasets = []
    if n_clusters is None:
        for cluster_no in range(2, 15):
            datasets.append((data, {"name": f"CS_{cluster_no}", "n_clusters": cluster_no}))
    else:
        datasets = [(data, {"name": f"CS_{n_clusters}", "n_clusters": n_clusters})]

    if selected_clusters is None:
        selected_clusters = ["Ward", "Birch", "SpectralClustering", "RawKMeans", "GaussianMixture", "Avg"]

    if default_base is None:
        default_base = {
            "quantile": 0.2,
            "eps": 0.3,
            "damping": 0.5,
            "preference": 5,
            "n_neighbors": 5,
            "n_clusters": 3,
            "min_samples": 5,
            "xi": 0.05,
            "min_cluster_size": 0.1,
            "name": "",
            "birch_th": 1.7,
            "affinity": "cosine",
            "fuzziness": 1.2,
        }

    all_outputs = pd.DataFrame(index=samples)

    # Run the experiments
    results = clustering_methods(datasets, default_base, samples, selected_clusters)

    models = []
    all_metrics = pd.DataFrame()
    for exp_name, output, cluster_models in results:
        # added the aditional data
        algorithm_names = output.iloc[:, 1:].columns

        if exp_name not in output.columns[1]:
            new_cols = {col_name: col_name + "_" + exp_name for col_name in algorithm_names}
            new_cols["samples"] = "Samples"
            output.rename(columns=new_cols, inplace=True)

        models.append((exp_name, cluster_models))
        all_outputs = pd.concat([all_outputs, output.set_index("Samples")], axis=1)

        # plot metrics
        metrics_df = compute_cluster_metrics(cluster_models, data)
        if 0:
            plot_cluster_metrics(metrics_df, exp_name)

        metrics_df["Exp"] = exp_name
        metrics_df["Cluster"] = metrics_df["Cluster"] + "_" + exp_name
        all_metrics = pd.concat([all_metrics, metrics_df])

    # make the biggest cluster is 0
    # pca_bu_df = pca_bu_df.sort_values(by="Sample").reset_index(drop=True)
    pca_df = pd.concat([pca_df.set_index("Sample"), all_outputs], axis=1)
    all_outputs.reset_index().rename(columns={"index": "Samples"})

    for col in all_outputs.columns:
        pca_df[col] = order_labels_size(pca_df[col])
        all_outputs[col] = order_labels_size(pca_df[col])

    all_outputs = pca_df.copy(deep=True)

    if show_figures:
        pca_df = pca_df.reset_index().rename(columns={"index": "Sample"})
        # add_metadata(pca_bu_df, dummy_meta, type="TCGA")
        plot_cols = list(pca_df.columns[2:-7])

        if show_consensus:
            plot_cols = ["TCGA408_classifier"] + plot_cols  # "consenus" add consensus
            pca_df["consenus"] = pd.factorize(pca_df["2019_consensus_classifier"])[0]
            pca_df["TCGA408_classifier"] = pd.factorize(pca_df["TCGA408_classifier"])[0]
            pca_df["consenus"] = order_labels_size(pca_df["consenus"])
            pca_df["TCGA408_classifier"] = order_labels_size(pca_df["TCGA408_classifier"])

        if rob_comp is not None:
            # transform labels to codes
            for col in rob_comp.columns:
                rob_comp[col] = order_labels_size(rob_comp[col])
                plot_cols.append(col)
            pca_df = pd.concat([pca_df, rob_comp], axis=1)

        if len(plot_cols) / 9 > 1:
            no_plots = int(len(plot_cols) / 9)

            for no_plot in range(0, no_plots + 1):
                fig = plot_clusters(
                    pca_df,
                    plot_cols[no_plot * 9 : (no_plot + 1) * 9],
                    x_label="PC_1",
                    y_label="PC_2",
                    exp_name="Various cluster methods I",
                    hover_data=list(pca_df.columns[-7:]),
                    pca_variance=pca_plot.explained_variance_ratio_,
                )
                fig.show()
        else:
            fig = plot_clusters(
                pca_df,
                plot_cols,
                x_label="PC_1",
                y_label="PC_2",
                exp_name="Various cluster methods",
                hover_data=list(pca_df.columns[-7:]),
                pca_variance=pca_plot.explained_variance_ratio_,
            )
            fig.show()

    # add metadata
    all_outputs = all_outputs.reset_index().rename(columns={"index": "Sample"})

    return all_outputs, models, all_metrics, pca_model


def compute_cluster_metrics(cluster_models, data, id=""):
    """Calculates the ‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’, ‘manhattan’ metrics for a given set of cluster models

    Args:
        cluster_models ([list]): List of cluster models such as Kmeans, Birch, Ward etc.
        data ([DataFrame]): The data used to trained the cluster models
        id ([String]): The id to append to the cluster names. This is useful for a set of experiments. Defaults to ""

    Returns:
        [DataFrame]: DataFrame used for storing the cluster metrics
    """

    clusters_metrics, cluster_names, sample_metrics = [], [], []
    for name, cluster_model in cluster_models:
        if name not in ["GaussianMixture", "FuzzyCMeans"]:
            labels = cluster_model.labels_
        else:
            labels = cluster_model.predict(data.values)

        # only if there is more than 1 label
        if len(set(labels)) > 1:
            # ‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’, ‘manhattan’

            sample_sillhouette_cosine = metrics.silhouette_samples(data, labels, metric="cosine")

            silhoute_metric = np.mean(sample_sillhouette_cosine)
            silhoute_metric_2 = metrics.silhouette_score(data, labels, metric="manhattan")
            silhoute_metric_3 = metrics.silhouette_score(data, labels, metric="cosine")
            calinski_harabasz_metric = metrics.calinski_harabasz_score(data, labels)
            davies_bouldin_metric = metrics.davies_bouldin_score(data, labels)

            sample_sillhouette_cosine = metrics.silhouette_samples(data, labels, metric="cosine")
            clusters_metrics.append([silhoute_metric, silhoute_metric_2, silhoute_metric_3, calinski_harabasz_metric, davies_bouldin_metric])
            sample_metrics.append((name + id, sample_sillhouette_cosine))
        else:
            clusters_metrics.append([0, 0, 0, 0, 0])
        cluster_names.append(name + id)

    metrics_names = ["Silhoute_euclidean", "Silhoute_manhattan", "Silhoute_cosine", "Calinski_habrasz", "Davies_bouldin"]
    cluster_metrics = pd.DataFrame(np.around(clusters_metrics, decimals=5), columns=metrics_names)
    cluster_metrics.insert(0, "Cluster", cluster_names)

    cluster_metrics["Samples_Silhouete_cos"] = sample_metrics

    return cluster_metrics


def apply_pca(df, pca_components, transpose=True, samples_col="samples", genes_col="genes"):
    """For a given DataFrame we apply Principal Component Analysis

    Args:
        df (DataFrame): The data to be processed
        samples_names ([list]): Neeed to add back to the returned df
        genes_names ([pd.Series]): Neeed to add back to the returned df
        pca_components ([int]): number of components
        transpose (bool, optional): This is needed so that we know how to add the genes/samples back to the df. Defaults to True.

    Returns:
        [DataFrame]: The PCA with samples and genes
    """
    # apply PCA
    samples_names = df.index.values
    cols = df.columns.values
    pca = PCA(n_components=pca_components)
    pca_bu = pca.fit_transform(df)
    print("Variation per principal component {} and the sum {:.02f}%".format(pca.explained_variance_ratio_, np.sum(pca.explained_variance_ratio_) * 100))

    # generate columns labels
    pca_col_labels = []
    for idx in range(pca_components):
        pca_col_labels.append("PC_" + str(idx + 1))  # start from 1

    pca_bu_df = pd.DataFrame(data=pca_bu, columns=pca_col_labels)
    # add dataframes labels accordingly
    dummy_df = df.copy(deep=True)
    if transpose:
        dummy_df.columns = cols
        dummy_df.insert(0, samples_col, samples_names)
        pca_bu_df.insert(0, samples_col, samples_names)
    else:
        dummy_df.columns = samples_names
        dummy_df.insert(0, genes_col, cols)
        pca_bu_df.insert(0, genes_col, cols)

    return pca_bu_df, pca


# plotting functions
def plot_clusters(df, plot_cols, x_label, y_label, exp_name, hover_data=None, marker_text=None, marker_text_position=None, pca_variance=None):
    """Function that plots the results from the clustering methods. It can received any number of columns values strings and it will display 3 on a row. This means that it will always display plots with 3 columns

    Args:
        df ([DataFrame]): At the moment only it supports 2d result DataFrames. This means can be either a t-sne or PCA
        plot_cols ([list]): A list of strings with the columns values to be ploted
        x_label([String]): The column of x which should be plotted, for t-sne it can be sne_2d_one. It has to be in df arg
        y_label ([String]): The column of y which should be plotted, for t-sne it can be sne_2d_two. It has to be in df arg
        colours_pool ([list]): The list of colours which are used for plotting
        exp_name ([String]): Name for experiment which is used for plots
        marker_text ([String], optional): The string which is a column name from the df. Defaults to None.
        marker_text_position ([String], optional): A string identified form plotly that's used to center accordingly the marker text. Defaults to None.
    """
    num_cols = 3
    num_rows = int(np.ceil(len(plot_cols) / num_cols))
    if hover_data == None:
        hover_data = plot_cols[:]
        hover_data.append("samples")

    fig = make_subplots(
        rows=num_rows, cols=num_cols, subplot_titles=plot_cols, horizontal_spacing=0.1, vertical_spacing=0.1, shared_xaxes=True, shared_yaxes=True
    )

    traces = []
    for plot_col in plot_cols:
        hover_data.append(plot_col)
        trace = trace_2d(
            df, plot_col, x_label, y_label, hover_cols=hover_data, dot_colour=plot_col, marker_text=marker_text, marker_text_position=marker_text_position
        )
        traces.append(trace)
        hover_data.remove(plot_col)

    # add the traces to the subplot
    subplot_titles = []
    idx_row, idx_col = 1, 1
    for trace in traces:
        subplot_titles.append(trace["name"])
        fig.add_trace(trace, row=idx_row, col=idx_col)
        # What we do here, we increment the column and row idxs. This means we increment the column at each iteration and reset it when it is divisible with the max number of columnbs. After that we increment the row idx
        if idx_col % num_cols == 0:
            idx_col = 0
            idx_row += 1
        idx_col += 1

    if pca_variance is not None:
        x_label = x_label + " ({:.02f}%)".format(pca_variance[0] * 100)
        y_label = y_label + " ({:.02f}%)".format(pca_variance[1] * 100)

    layout = go.Layout(title_text=exp_name)

    fig.update_traces(marker=dict(size=12, line=dict(width=1, color="Black")), selector=dict(mode="markers"))
    fig.update_xaxes(title_text=x_label)
    fig.update_yaxes(title_text=y_label)
    fig.update_layout(layout, height=1500)
    return fig


def trace_2d(df, title, x_axis, y_axis, hover_cols=None, dot_colour=None, marker_text=None, marker_text_position=None):
    """This creates a 2D trace that it can be later be used for subplots in plotly. Easy to customise, the only constriant is that the x and y axis has to be part of the df argument.

    Args:
        df ([DataFrame]): The data that contains the points to be plotted.
        colours_pool ([list]): The list of colours which are used for plotting
        title ([String]): The title of the plot
        x_axis ([String]): X label, it has to be part of the DataFrame
        y_axis ([String]): Y label, it has to be part of the DataFrame
        hover_cols ([list], optional): [description]. Defaults to (centroids and samples for backwards compatibility).
        dot_colour ([String], optional): The string which is the column name of the df and by that it's done the colouring. Defaults to centroids (for backwards compatibility).
        marker_text ([String], optional): The string which is a column name from the df. Defaults to None.
        marker_text_position ([String], optional): A string identified form plotly that's used to center accordingly the marker text. Defaults to None.

    Returns:
        [trace]: The trace object that was created
    """
    hover, colours, markers = [], [], {}
    mode_markers = "markers"
    text_list = []
    if hover_cols == None:
        hover_cols = ["centroids", "samples", "diagnosis", "labels_ter_avg", "labels_ter_sd"]
    if dot_colour == None:
        dot_colour = "centroids"
    if marker_text != None:
        mode_markers = "markers+text"
        text_list = df[marker_text].values
        if marker_text_position == None:
            marker_text_position = "bottom left"
    for _, row in df.iterrows():
        centroid = row[dot_colour]
        # create the hover data
        hover_string = ""
        for hover_col in hover_cols:
            hover_string += "<br>%s=%s" % (hover_col, str(row[hover_col]))
        hover_string += "<br>" + x_axis + "=%{x}" + "<br>" + y_axis + "=%{y}"
        hover.append(hover_string)

        colours.append(colours_pool[centroid])

    markers["color"] = colours
    markers["size"] = 6
    trace = dict(
        type="scatter",
        x=df[x_axis],
        y=df[y_axis],
        hovertemplate=hover,
        showlegend=True,
        name=title,
        marker=markers,
        mode=mode_markers,
        text=text_list,
        textposition=marker_text_position,
    )
    return trace


def plot_cluster_metrics(metrics, exp_name, hide_text=False):
    """Creates the the figures for a given DataFrame of metrics.

    Args:
        metrics (DataFrame): Holds the metrics values, columns names being represented by the algorithm for which the metrics
        algorithm_names ([list]): List of strings of the algorithms that are being run
        exp_name ([String]): Name of the experiment
        hide_text ([Bool]): If True the points text is hidden. This is usefull when there are lots of points. Defaults to False
    """
    fig = go.Figure()
    # define x axis
    metrics_names = ["Silhoute_euclidean", "Silhoute_manhattan", "Silhoute_cosine", "Calinski_habrasz", "Davies_bouldin"]
    mode, text, traces = "lines+markers+text", metrics["Cluster"].values, []
    random_x = np.linspace(0, 15, metrics.shape[0])
    if hide_text:
        mode = "lines+markers"
    for metrics_name in metrics_names:
        trace = go.Scatter(x=random_x, y=metrics[metrics_name], mode=mode, text=text, hoverinfo="all", textposition="top right")
        traces.append(trace)

    fig = make_subplots(
        rows=3,
        cols=2,
        subplot_titles=[
            "Silhouette euclidean (higher the better)",
            "Silhouette manhattan (higher the better)",
            "Silhouette cosine (higher the better)",
            "Calinski-Harabrasz (higher the better)",
            "Davies-Bouldin (lower the better)",
        ],
        shared_yaxes=False,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
    )

    for idx, trace in enumerate(traces):
        fig.add_trace(trace, row=int(idx / 2) + 1, col=idx % 2 + 1)

    layout = go.Layout(title_text="Different clustering metrics " + exp_name)
    fig.update_layout(layout)
    fig.show()


def mark_best_3(ref, metrics, metric="Silhoute_cosine"):
    # Remove the experiments w/ CS_2 and CS_3
    ascending = False
    if metric == "Davies_bouldin":
        ascending = True
    best_sill = metrics[~metrics["Exp"].str.contains("2|3")].sort_values(by=[metric], ascending=ascending)[:3]

    offset = best_sill[metric].mean() / 50

    annotations = []
    i = 0
    for i_dx, row in best_sill.iterrows():
        i += 1
        for j in range(1, i + 1):
            annotations.append(
                dict(x=row["Cluster"], y=row[metric] + offset * i * j, text="x", showarrow=False, xref="x{}".format(ref), yref="y{}".format(ref))
            )
    return annotations


def display_metrics(metrics, exp_title="Metrics", show_individual=False, verbose=True):
    """
        Receives the Metrics DataFrame from comp_exp function above.

    Args:
        metrics ([DataFrame]): It needs to contain
        exp_name (str, optional): [description]. Defaults to "".
    """
    # other sillhouettes available (Silhoute_euclidean', 'Silhoute_manhattane')

    metrics["cluster_type"] = ["-".join(cluster.split("_")[:1]) for cluster in metrics["Cluster"]]

    title_1 = "Cosine Sillhouette (higher the better)"
    fig_1 = px.bar(metrics, y="Silhoute_cosine", x="Cluster", title="{}. {}".format(exp_title, title_1), color="cluster_type")
    fig_1.update_yaxes(title_text="Sillhouette Avg")
    fig_1.update_xaxes(title_text="Experiment")
    annotations_1 = mark_best_3(1, metrics, metric="Silhoute_cosine")

    title_2 = "Calinski_habrasz (higher the better)"
    fig_2 = px.bar(metrics, y="Calinski_habrasz", x="Cluster", title="{}. {}".format(exp_title, title_2), color="cluster_type")
    fig_2.update_yaxes(title_text="Calinski Habrasz Avg")
    fig_2.update_xaxes(title_text="Experiment")
    fig_2.update_layout(xaxis={"categoryorder": "array", "categoryarray": metrics["Cluster"]})
    annotations_2 = mark_best_3(2, metrics, metric="Calinski_habrasz")

    title_3 = "Davies_bouldin (lower the better)"
    fig_3 = px.bar(metrics, y="Davies_bouldin", x="Cluster", title="{}. {}".format(exp_title, title_3), color="cluster_type")
    fig_3.update_yaxes(title_text="Davies Bouldin Avg")
    fig_3.update_xaxes(title_text="Experiment")
    fig_3.update_layout(xaxis={"categoryorder": "array", "categoryarray": metrics["Cluster"]})
    annotations_3 = mark_best_3(3, metrics, metric="Davies_bouldin")

    # Extract the traces
    figs = [fig_1, fig_2, fig_3]
    all_traces = []
    for fig in figs:
        fig_traces = []
        for trace in range(len(fig["data"])):
            fig_traces.append(fig["data"][trace])
        all_traces.append(fig_traces)

    subplots = make_subplots(rows=3, cols=1, subplot_titles=[title_1, title_2, title_3], shared_xaxes=True, vertical_spacing=0.05)

    for idx, fig_traces in enumerate(all_traces):
        for traces in fig_traces:
            subplots.append_trace(traces, col=1, row=idx + 1)

    # the subplot as shown in the above image
    layout = go.Layout(title_text="{}. Combined metrics ".format(exp_title))
    subplots.update_layout(layout, annotations=annotations_1[:3] + annotations_1 + annotations_2 + annotations_3)

    visited = []
    for trace in subplots["data"]:
        trace["name"] = trace["name"].split(",")[0]
        if trace["name"] not in visited:
            trace["showlegend"] = True
            visited.append(trace["name"])
        else:
            trace["showlegend"] = False

    if show_individual:
        fig_1.show()
        fig_2.show()
        fig_3.show()

    return subplots


def plot_meta_scores(df, y_axis="IFNG", classification="Basal_type", size="infiltration_score"):
    # infg_df["2019CC_Ba/Sq"] = infg_df["2019CC_Ba/Sq"].astype(float)
    # infg_df["2019CC_Stroma-rich"] = infg_df["2019CC_Stroma-rich"].astype(float)

    category_order = {
        "Basal_type": ["IFNG-IFNG (High) - 42", "Basal-IFNG (Medium) - 38", "Basal-Basal (Low) - 40"],
        "infiltration_label": ["High Inf", "Medium Inf", "No Inf"],
    }

    title = "Classification by {}. IFNG vs infiltration score; Colouring by Estimate score".format(classification)
    fig1 = px.scatter(df, x="ESTIMATE_score", y=y_axis, color=classification, title=title, category_orders=category_order, size=size)
    fig1.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
    # fig1.show()

    title = "Classification by {}. IFNG vs infiltration score; Colouring by Imune rich score".format(classification)
    fig = px.scatter(df, x="Immune_score", y=y_axis, color=classification, title=title, category_orders=category_order, size=size)

    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
    # fig.show()

    title = "Classification by {}. IFNG vs infiltration score; Colouring by Stroma rich score".format(classification)
    fig = px.scatter(df, x="Stromal_score", y=y_axis, color=classification, title=title, category_orders=category_order, size=size)

    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
    # fig.show()

    return fig1


# Helper functions
def order_labels_size(df_series):
    """This function ensures that the the cluster labels are always in the order of the cluster sizes. This means that the cluster label 0 will correspond for the largest cluster and for n-clusters the n-1 will be the cluster with the lowest members.

    Args:
        df_series (Pandas Series): The cluster labels to be mapped

    Returns:
        [Pandas Series]: New mappped pandas series
    """
    # ordered by the number of frequency
    cluster_count = [clust for clust, _ in Counter(df_series).most_common()]
    new_labels = list(range(len(df_series.unique())))
    dic = dict(zip(cluster_count, new_labels))
    # print("Remap of the old vs new labels",dic)
    return df_series.map(dic)


def elbow_method(df, min_k, max_k, label):
    """Apply elbow method on K-means for a given data (DataFrame)

    Args:
        df ([DataFrame]): Input data
        min_k ([Int]):
        max_k ([Int]):

    Returns:
        [plotly.Figure, list]: Return the figure to be plotted and the sum squared distances
    """
    sum_dist, K = [], range(min_k, max_k)
    for k in K:
        km = cluster.KMeans(n_clusters=k, n_init=10)
        km = km.fit(df)
        sum_dist.append(km.inertia_)

    data = go.Scatter(x=list(K), y=sum_dist, mode="lines+markers")
    layout = go.Layout(
        title=f"{label}. Elbow method based on K-means inertia (sum sqrd dist)",
        xaxis=go.layout.XAxis(title="Clusters"),
        yaxis=go.layout.YAxis(title="Sum of squared distances"),
        title_y=0.9
    )
    fig = go.Figure(data=data, layout=layout)


    return fig, sum_dist



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   PGCNA_output.py
@Time    :   2023/02/21 15:54:50
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   Don't be lazy fill this in!
"""

import os

import igraph as ig
import numpy as np
import pandas as pd
import plotly.express as px
from scipy.stats import zscore
import re

from .utilities import clustering as cs


class NetworkOutput:
    name: str
    exp_meta: pd.DataFrame
    type: str
    graph: ig.Graph
    meta_path: str
    pgcna_path: str
    graph_type: str
    leiden_top3: pd.DataFrame
    meta_df: pd.DataFrame

    nodes_df: pd.DataFrame
    edges_df: pd.DataFrame
    tpm_df: pd.DataFrame
    mut_tf: pd.DataFrame

    def __init__(self, graph, exp_meta, mut_df, exps_path, rel_path="", exp_type="PGCNA", **kwargs):
        self.exps_path = exps_path
        self.exp_meta = exp_meta

        # choosing the experiment
        if exp_type == "PGCNA":
            self.init_pgcna(graph, exp_meta, mut_df, rel_path)
        else:
            self.init_inet(graph, exp_meta, mut_df, base_path=kwargs["base_path"])
        self.exp_type = exp_type

    def init_pgcna(self, graph, exp_meta, mut_df, rel_path):
        self.name = exp_meta["expName"]
        exp_comp = self.name.split("_")
        genesKept = int(int(exp_meta["genesKept"]) / 1000)
        self.type = f"{exp_comp[0]}_{genesKept}K"
        self.graph = graph
        self.graph_type = "ig"
        exp_meta["edgesPG"] = int(exp_meta["edgesPG"])
        exp_meta["edgesTF"] = int(exp_meta["edgesTF"])

        # check if the exps diviate from the 'standard' values
        self.type = f"{self.type}_{exp_meta['edgesTF']}TF"
        if exp_meta["edgesPG"] != 3:
            self.type = f"{self.type}_{exp_meta['edgesPG']}EPG"

        if "resolution" in exp_meta.index:
            self.type = f"{self.type}_{exp_meta['resolution']}res"

        # temporary
        if rel_path == "":
            files_used = exp_meta["filesUsed"]
        else:
            files_used = os.path.abspath(os.path.join(rel_path, exp_meta["filesUsed"]))

        # setting the path
        self.meta_path = f"/CORR_MATRIX_GMF0/{files_used.split('/')[-1]}_RetainF{exp_meta['retainF']}_metaInf.txt"
        self.pgcna_path = f"{self.exps_path}/PGCNA/{self.name}/EPG{exp_meta['edgesPG']}/"

        ###### loading dataframes ######
        #  Leiden
        leidenalg_master = pd.read_csv(self.pgcna_path + "/LEIDENALG/moduleInfoBest.txt", sep="\t")
        self.leiden_top3 = leidenalg_master.iloc[:3].copy(deep=True)

        #  get all the tpms used in that run
        all_tpms = pd.read_csv(files_used, sep="\t", index_col="genes")
        #  meta df
        self.meta_df = pd.read_csv(self.pgcna_path + self.meta_path, sep="\t", header=None)

        # processing
        #  rename meta columns
        self.meta_df.columns = ["genes", "q2E", "varWithin", "varAcross"]
        #  only the best run of leiden
        best_mod = str(leidenalg_master.iloc[0]["Mod#"]) + ".csv"
        self.leiden_best = pd.read_csv(self.pgcna_path + "/LEIDENALG/BEST/ClustersTxt/" + best_mod)
        #  only the genes ones used in PGCNA
        self.tpm_df = all_tpms.loc[self.leiden_best["Id"]]
        #  mut_df
        dmy_df = pd.DataFrame(
            data=mut_df.loc[mut_df.index.isin(self.leiden_best["Id"])],
            index=self.leiden_best["Id"],
        )
        dmy_df.fillna(0, inplace=True)
        self.mut_df = dmy_df

    def init_inet(self, graph: ig.Graph, exp_meta: pd.Series, mut_df: pd.DataFrame, base_path: str):
        # Same as for PGCNA parsing
        self.name = exp_meta["expName"]
        genesKept = int(int(exp_meta["genesKept"]) / 1000)
        self.type = f"{exp_meta['modifierType']}_{genesKept}K"
        self.graph = graph
        self.graph_type = "ig"
        exp_meta["edgesPG"] = int(exp_meta["edgesPG"])
        exp_meta["edgesTF"] = int(exp_meta["edgesTF"])
        self.base_path = base_path

        self.type = f"{self.type}_{exp_meta['edgesTF']}TF"
        # TODO do we need this?
        if exp_meta["edgesPG"] != 3:
            self.type = f"{self.type}_{exp_meta['edgesPG']}EPG"

        if ("resolution" in exp_meta.index) and (not np.isnan(exp_meta["resolution"])):
            self.type = f"{self.type}_{exp_meta['resolution']}res"

        if "sbm_method" in exp_meta.index and not pd.isnull(exp_meta["sbm_method"]):
            self.type = f"{self.type}_{exp_meta['sbm_method']}"
            self.sbm_method = exp_meta["sbm_method"]
            self.graph_type = "gt"

        # This is *different* from the PGCNA parsing
        ### Defining the file paths
        tpm_path = os.path.abspath(os.path.join(base_path, exp_meta["input_ge_file"]))
        self.pgcna_path = f"{self.exps_path}/Networks/{self.name}/EPG{exp_meta['edgesPG']}/"
        self.meta_path = f"{self.pgcna_path}/gene_stats.tsv"
        leiden_best_path = f"{self.pgcna_path}/Leiden/Best/"

        ###### loading dataframes ######
        #  Leiden
        leidenalg_master = pd.read_csv(self.pgcna_path + "/Leiden/summary_leiden.tsv", sep="\t")
        self.leiden_top3 = leidenalg_master.iloc[:3].copy(deep=True)

        #  get all the tpms used in that run
        all_tpms = pd.read_csv(tpm_path, sep="\t", index_col="gene")

        #  meta df
        self.meta_df = pd.read_csv(self.meta_path, sep="\t")

        ####### processing
        #  rename meta columns; this is was needed for PGCNA parsing. Kept it for consistency
        self.meta_df.columns = ["genes", "q2E", "varWithin", "varAcross"]
        #  Select the best run
        best_mod = f"{leidenalg_master.iloc[0]['Mod#']}.tsv"
        self.leiden_best = pd.read_csv(f"{leiden_best_path}/leiden_best_{best_mod}", sep="\t", index_col="Gene")

        # align with the PGCNA parsers
        all_tpms.index.names = ["genes"]
        self.meta_df.index.names = ["genes"]

        self.leiden_best.index.names = ["Id"]
        self.leiden_best.reset_index(inplace=True)
        #### TPMs
        # only the genes ones used in PGCNA
        self.tpm_df = all_tpms.loc[self.leiden_best["Id"]]
        #### Mutations
        dmy_df = pd.DataFrame(
            data=mut_df.loc[mut_df.index.isin(self.leiden_best["Id"])],
            index=self.leiden_best["Id"],
        )
        dmy_df.fillna(0, inplace=True)
        self.mut_df = dmy_df

    #### PGCNA scores ####
    def compute_scores(self):
        """

        Gets both MEV and ModCon scores to which the graph stats are added.

        Args:
            exp (dict): information about the experiments

        Returns:
            df, df, list: _description_
        """
        # a dictionary with key - the idx of the community and the value the ModCon

        modCons = self.get_ModCon()
        mevs = self.get_mevs(self.tpm_df, modCons, num_genes=50)

        self.graph_stats = self.compute_graph_stats()
        success = True
        for key in modCons.keys():
            genes = modCons[key].index.values
            modCons[key] = pd.concat([modCons[key], self.mut_df["count"], self.graph_stats], axis=1).dropna()
            if modCons[key].shape[0] != len(genes):
                print(f"❌ Failed data merger for (Mod {key})!")
                success = False

        if not success:
            print(f"❌ There was an error in merging the data for {self.name} ")
            raise ValueError("Problem in merging the data")

        self.modCons = modCons
        self.mevs = mevs

    def get_ModCon(self):
        leiden_best = self.leiden_best
        mut_df = self.mut_df
        meta_df = self.meta_df

        # create a DataFrame with the components for the modCon equation
        modifier = self.type.split("_")[0]
        col = "conn_{}".format(modifier)

        # all_genes = []  # can be used for testing
        modCons = {}
        for mod_class in leiden_best["Modularity Class"].unique():
            genes = leiden_best.loc[leiden_best["Modularity Class"] == mod_class]["Id"].values

            conn_g = []
            for gene in genes:
                # 1st condition, get all the edges that the current gene is the Source and are inside the module
                # 2nd condition, get all the edges that current is the Target and are inside the module
                in_edges = self.edges_df[(self.edges_df["Source"] == gene) | (self.edges_df["Target"] == gene)]
                in_edges = in_edges[in_edges["Source"].isin(genes)]
                in_edges = in_edges[in_edges["Target"].isin(genes)]
                weights_sum = in_edges['Weight'].sum()
                if weights_sum == 0:
                    print(f"Weighted sum = 0 for Com {mod_class} !!")
                conn_g.append([gene, weights_sum])

            conn_df = pd.DataFrame(conn_g, columns=["gene", col]).set_index("gene")
            working_df = pd.concat([conn_df, meta_df.loc[meta_df["genes"].isin(genes)].set_index("genes"), mut_df[mut_df.index.isin(genes)]["count"]], axis=1)

            # 5. Workout the ModCon and save it
            working_df["ModCon_{}".format(self.type)] = (
                (working_df[col] ** 2) * working_df["q2E"] * working_df["varWithin"] * (100 - working_df["varAcross"]) / 100
            )

            # sort by the current exp type
            modCons[mod_class] = working_df.sort_values(by=["ModCon_{}".format(self.type), "varAcross"], ascending=[False, False])

        self.modCons = modCons
        return modCons

    def get_mevs(self, tpms, modCon, sort_col="ModCon", num_genes=25, verbose=False):
        mevs = pd.DataFrame(index=tpms.columns)
        tpms_log = np.log2(tpms + 1)
        info = {}  # used for verbose
        for key, value in modCon.items():
                
            data = value.sort_values(by=sort_col, ascending=False).iloc[:num_genes].copy(deep=True)
            genes = data.index.values
            df = tpms_log[tpms_log.index.isin(genes)].transpose()

            # 1. Per gene, standardize (z-score) the quantile normalized log2 TPM
            for gene in df.columns:
                tst = (df[gene] - df[gene].min()) / (df[gene].max() - df[gene].min())
                # tst = percentileofscore(tst, tst)
                df[gene] = zscore(tst)

            # 2.Sum up the z-scores to give the mev
            mevs[f"Com_{key}"] = df.sum(axis=1)

            # Useful information
            not_found, matched = [], []
            if len(genes) != df.shape[1] and verbose:
                not_found = list(set(genes) - set(df.columns.values))
                matched = list(set(genes) & set(df.columns.values))
                # print("Comm {}. Genes not found {}/{}({:.2f}%)".format(key, not_found, len(genes), not_found/len(genes)*100)
            else:
                matched = list(genes)

            if df.shape[1] == 0:
                print(f"No genes matched in Com {key}")
                continue

            if verbose:
                info[key] = {"modCon_genes": genes, "matched": matched, "not_matched": not_found, "mevs": df.sum(axis=1)}

        return mevs, info

    def get_iMevs(self, h_tpms: pd.DataFrame, tum_tpms: pd.DataFrame, modCon: dict, sort_col="ModCon", num_genes=25, verbose=False, **kwargs):

        # Integrated mevs
        i_mevs = pd.DataFrame(index=tum_tpms.columns) 

        # Mevs for the same dataset
        # t_mevs = pd.DataFrame(index=h_tpms.columns)
        t_log2 = np.log2(tum_tpms + 1)
        h_log2 = np.log2(h_tpms + 1)

        # Check if there are mutations
        mut_offset = 0
        if "mut_df" in kwargs.keys():
            mut_offset = 1
            mut_df = kwargs.pop("mut_df")

            if "offset" in kwargs.keys():
                mut_offset = kwargs.pop('offset')

            print(f"### Running with mut_ofset {mut_offset}")

        info = {}  # used for verbose
        for key, value in modCon.items():
            data = value.sort_values(by=sort_col, ascending=False).iloc[:num_genes].copy(deep=True)
            genes = data.index.values

            # find the genes in the dataset
            df_int = t_log2[t_log2.index.isin(genes)].transpose()

            # Neeed to compute the stats
            df_h = h_log2[h_log2.index.isin(genes)].transpose()

            # Only use the genes common
            cmn_genes = set(df_int.columns) & set(genes)
            diff_genes = set(genes) - set(df_int.columns)

            # 1. Per gene, standardize (z-score) the normalized log2 TPM
            for gene in cmn_genes:
                # Normalised and get the precentiles for both datasets
                h_norm = (df_h[gene] - df_h[gene].min()) / (df_h[gene].max() - df_h[gene].min())
                t_norm = (df_int[gene] - df_int[gene].min()) / (df_int[gene].max() - df_int[gene].min())
                # Stats
                h_mean, h_std = h_norm.mean(), h_norm.std()

                # z-scores
                ## Integrated Mevs
                df_int[gene] = (t_norm - h_mean) / h_std

            ## Adding the representation of the mutations
            if mut_offset != 0:
                sel_df = mut_df.loc[list(cmn_genes)]
                df_int = df_int + sel_df.T * mut_offset

            # 2.Sum up the z-scores to give the mev
            i_mevs[f"Com_{key}"] = df_int.sum(axis=1)

            # Useful information
            not_found, matched = [], []
            if len(genes) != df_int.shape[1] and verbose:
                not_found = list(set(genes) - set(tum_tpms.columns.values))
                matched = list(set(genes) & set(tum_tpms.columns.values))

            if df_h.shape[1] == 0:
                print(f"No genes matched in Com {key}")
                continue

            if verbose:
                info[int(key)] = {
                    "modCon_genes": genes,
                    "matched": matched,
                    "not_matched": not_found,
                    "diff_genes": diff_genes,
                    "cmn_genes": cmn_genes,
                    "h_norm": h_norm,
                    "t_norm": t_norm,
                }

        return i_mevs, info

    ### Connectivity ####
    def get_connectivity(self):
        leiden_best = self.leiden_best

        # all_genes = []  # can be used for testing
        all_conn_mod = dict()
        for mod_class in leiden_best["Modularity Class"].unique():
            genes = leiden_best.loc[leiden_best["Modularity Class"] == mod_class]["Id"].values

            conn_module = []
            for gene in genes:
                in_edges = self.edges_df[
                    ((self.edges_df["Source"] == gene) & (self.edges_df["Target"].isin(genes)))
                    | ((self.edges_df["Target"] == gene) & (self.edges_df["Source"].isin(genes)))
                ]["Target"]
                out_edges = self.edges_df[(self.edges_df["Source"] == gene) | (self.edges_df["Target"] == gene)]["Target"]
                conn_module.append([gene, in_edges.shape[0], out_edges.shape[0], list(in_edges.values), list(out_edges.values)])

            all_conn_mod[mod_class] = conn_module

            all_df = pd.DataFrame()

        # put them in a dataframe
        for key, val in all_conn_mod.items():
            new_df = pd.DataFrame(val, columns=["gene", "in_edges", "out_edges", "in_genes", "out_genes"])
            new_df["comm"] = key
            all_df = pd.concat([all_df, new_df], axis=0)

        self.genes_conn_edges = all_df

        return all_df

    ### Network stats ####
    def ge_comm(self):
        comm_tpms = {}
        median_vals, std_df = pd.DataFrame(index=self.tpm_df.columns), pd.DataFrame(index=self.tpm_df.columns)

        for mod_class in self.leiden_best["Modularity Class"].unique():
            genes = self.leiden_best.loc[self.leiden_best["Modularity Class"] == mod_class]["Id"].values
            comm_tpms[mod_class] = self.tpm_df.loc[self.tpm_df.index.isin(genes)]

            median_vals = pd.concat([median_vals, comm_tpms[mod_class].median().rename("Comm_{}".format(mod_class))], axis=1)
            std_df = pd.concat([std_df, comm_tpms[mod_class].std().rename("Comm_{}".format(mod_class))], axis=1)

        med_df = median_vals.melt(var_name="Comm", value_name="Median", ignore_index=False).reset_index(names=["Sample"])
        std_df = std_df.melt(var_name="Comm", value_name="Std", ignore_index=False).reset_index(names=["Sample"])
        combined_stats = pd.concat([std_df, med_df["Median"]], axis=1)
        return combined_stats

    #### Graph functions ####
    def compute_graph_stats(self):
        graph_stats = pd.DataFrame(index=self.graph.vs[:]["name"])
        graph_stats["betwenees"] = self.graph.betweenness(directed=False, weights="weight")
        graph_stats["closeness"] = self.graph.closeness(weights="weight")
        graph_stats["degree"] = self.graph.degree()
        graph_stats["strength"] = self.graph.strength()
        graph_stats["pageRank"] = self.graph.pagerank(directed=False, weights="weight")
        graph_stats["hubScore"] = self.graph.hub_score()

        return graph_stats

    def mergeSourceTarget(self, col_name):
        """
        Merge the 'Source' and 'Target' columns of the edges dataframe into a new 'Edge' column.
        Set the 'Edge' column as the index of the dataframe.
        Rename the 'Weight' column to the specified 'col_name'.

        Parameters:
            col_name (str): The name to be assigned to the renamed 'Weight' column.

        Returns:
            None
        """
        self.edges_df["Edge"] = self.edges_df["Source"] + "-" + self.edges_df["Target"]
        self.edges_df.set_index("Edge", inplace=True)
        self.edges_df.rename(columns={"Weight": col_name}, inplace=True)

    #### Export functions ####
    def export_to_gephi(self, save=False, sbm_df=None, state_idx=0, hsbm=False, **kwargs):
        # Build the edges and nodes DFs
        filename, path  = self.name, self.exps_path
        edge_list = []
        if self.graph_type == "ig":
            for edge in self.graph.es:
                source_vertex_id = edge.source
                target_vertex_id = edge.target
                source_vertex = self.graph.vs[source_vertex_id]
                target_vertex = self.graph.vs[target_vertex_id]
                weight = edge["weight"]
                edge_list.append((source_vertex["name"], target_vertex["name"], weight))

            nodes_df = pd.DataFrame([node["name"] for node in self.graph.vs], columns=["Id"])
        else:
            gene_list = list(self.graph.vp.gene)  # List of all gene names

            # Extract all edges with their weights
            edge_list = [(self.graph.vp.gene[edge.source()],
                    self.graph.vp.gene[edge.target()],
                    self.graph.ep.weight[edge])
                    for edge in self.graph.edges()]

            # Create DataFrames
            nodes_df = pd.DataFrame(gene_list, columns=["Id"])


        edge_df = pd.DataFrame(edge_list, columns=["Source", "Target", "Weight"])

        edge_df["Type"] = "undirected"
        edge_df["fromAltName"] = edge_df["Source"]
        edge_df["toAltName"] = edge_df["Target"]

        # Build the Nodes DataFrame
        nodes_df["Label"] = nodes_df["Id"]  # Gephi requirement(?)
        nodes_df["Gene"] = nodes_df["Id"]  # Needed to for the filter script
        nodes_df.set_index("Id", inplace=True)

        nodes_df = pd.concat(
            [
                nodes_df,
                self.leiden_best.set_index("Id")["Modularity Class"],
                self.mut_df["count"],
            ],
            axis=1,
        )

        # # TODO: No need for these now but need to update the lists
        # nodes_df = mk.addTF(nodes_df, gene_col="Label")
        top_genes = 100
        # if there is a modCon add the ranking
        if (self.graph_type == 'ig' and hasattr(self, "modCons")):
            for modCon, value in self.modCons.items():
                dmy = value.sort_values(by=["ModCon_{}".format(self.type)], ascending=False).reset_index(names="Id").iloc[:top_genes]
                dmy["Rank"] = dmy.index + 1
                dmy.set_index("Id", inplace=True)
                nodes_df.loc[nodes_df["Modularity Class"] == modCon, "ModCon_Rank"] = dmy["Rank"]
                # offset needed to make it easy to use in gephi
                nodes_df["ModCon_Rank"] = nodes_df["ModCon_Rank"].fillna(top_genes + 10)
        elif self.graph_type == 'gt': 
            # Add the communities information
            if "com_df" in kwargs.keys():
                nodes_df = pd.concat([nodes_df, kwargs.pop("com_df")], axis=1)

            # Add the modCon Rank
            if hasattr(self, "gt_modCon"):
                for modCon, value in self.gt_modCon.items():
                    dmy = value.sort_values(by=["ModCon_{}_gt".format(self.type)], ascending=False).reset_index(names="Id").iloc[:top_genes]
                    dmy["Rank"] = dmy.index + 1
                    dmy.set_index("Id", inplace=True)
                    nodes_df.loc[nodes_df["max_b"] == modCon, "ModCon_Rank"] = dmy["Rank"]
                    # offset needed to make it easy to use in gephi
                    nodes_df["ModCon_Rank"] = nodes_df["ModCon_Rank"].fillna(top_genes + 10)

        ### check if there is an IVI file
        ivi_path = f"{path}/PGCNA/{self.name}/IVI.tsv"
        if self.exp_type == "iNet":
            ivi_path = f"{path}/Networks/{self.name}/IVI.tsv"
        if os.path.exists(ivi_path):
            ivi_df = pd.read_csv(ivi_path, sep="\t")
            ivi_df.index.names = ["Id"]
            if self.exp_type == "iNet":
                nodes_df = pd.concat([nodes_df, ivi_df.reset_index().set_index("gene")[["IVI"]]], axis=1)
            else: 
                nodes_df = pd.concat([nodes_df, ivi_df], axis=1)

        if sbm_df is not None:
            if hsbm:
                # nodes_df = pd.concat([nodes_df, sbm_df], axis=1)
                # the concatenation gives some error in gephi
                for col in sbm_df.columns:
                    nodes_df[col] = sbm_df[col]
            else:
                nodes_df[f"SBM_{state_idx}"] = sbm_df["max_b"]
                nodes_df[f"SMB_stable_{state_idx}_num"] = sbm_df["stable_gene_num"]

        if hasattr(self, 'tf_list'):
            nodes_df["TF"] = 0
            nodes_df.loc[nodes_df.index.isin(self.tf_list), "TF"] = 1
        else:
            # Backwards compatibility with experiment control for all healthy
            tf_ctrl_path = "../../data/tf_ctrl.csv"
            if os.path.exists(tf_ctrl_path):
                genes = pd.read_csv(tf_ctrl_path).index
                nodes_df["ctrl_tf"] = 0
                nodes_df.loc[nodes_df.index.isin(genes), "ctrl_tf"] = 1

        if save:
            gephi_path = f"{path}/PGCNA/{filename}/"
            if self.exp_type == "iNet":
                gephi_path = f"{path}/Networks/{filename}/"

            
            if not os.path.exists(gephi_path):
                os.makedirs(gephi_path)

            print(f"Saving Gephi files to {os.path.abspath(gephi_path)}")

            nodes_df.index.names=['Id']
            nodes_df.to_csv(f"{gephi_path}/{filename}_nodes.tsv", sep="\t")
            edge_df.to_csv(
                f"{gephi_path}/{filename}_edges.tsv",
                sep="\t",
                index=False,
            )

        self.nodes_df = nodes_df
        self.edges_df = edge_df

    def export_morpheus_mevs(self, vu_output, name, tum=True, col='RawKMeans_CS_5', path=''):
        if hasattr(self, "mevsMut"):
            # Export if for morpheus
            mevs_name = f"mevs_{name}.tsv"

            # Morpheus doesn't like columns of strings
            dmy = None
            if tum:
                transform_cols = ["TCGA408_classifier", "consensus", "Lund2017.subtype"]
                vu_cols = []
                for col in transform_cols:
                    vu_output[col] = vu_output[col].astype("category")
                    vu_output[col + "_num"] = vu_output[col].cat.codes
                    vu_cols.append(col + "_num")

                vu_cols = vu_cols + ["RawKMeans_CS_5"]
                dmy = pd.concat(
                    [
                        vu_output[
                            [
                                "RawKMeans_CS_5",
                                "TCGA408_classifier_num",
                                "consensus_num",
                                "Lund2017.subtype_num",
                            ]
                        ],
                        self.mevsMut,
                    ],
                    axis=1,
                )

                # deal with missing samples between communities and VU_CS
                missing_vu = list(set(self.mevsMut.index) - set(vu_output.index))
                print(f"Missing #{len(missing_vu)} from VU_CS: {missing_vu}\n")
                dmy.loc[missing_vu, vu_cols] = -1

                missing_mevs = list(set(vu_output.index) - set(self.mevsMut.index))
                print(f"Missing #{len(missing_mevs)} from MEVs: {missing_mevs}\n")
                dmy.drop(index=missing_mevs, inplace=True)
            else:
                dmy = self.mevsMut


            save_path = f"{self.exps_path}stats/h_{mevs_name}.tsv"
            if path != '':
                save_path = f"{path}/{mevs_name}.tsv"

            dmy.transpose().to_csv(save_path, sep="\t", index=True)

            return dmy
        else:
            raise ValueError("No MEVS mutation")

    def prep_for_go(self, num_sel=None, to_save=False, save_path=None):
        top_genes = {}
        for mod_con in self.modCons:
            if num_sel:
                sel_genes = list(self.modCons[mod_con].sort_values(by="ModCon_{}".format(self.type), ascending=False).index.values[:num_sel])
            else:
                sel_genes = list(self.modCons[mod_con].sort_values(by="ModCon_{}".format(self.type), ascending=False).index.values[:])

            top_genes[mod_con] = [sel_genes, "\n".join(sel_genes)]
        dmy = pd.DataFrame.from_dict(top_genes, orient="index").reset_index(names="Community").rename(columns={0: "Genes", 1: "GO format"})

        if to_save and save_path:
            folder_path = "{}/GeneOntology/{}/".format(save_path, self.type)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)

            if num_sel:
                path = "{}/{}".format(folder_path, "h_comm_genes_{}.tsv".format(num_sel))
            else:
                path = "{}/{}".format(folder_path, "h_comm_genes.tsv")

            dmy.to_csv(path, index=False, sep="\t")

        return dmy

    @staticmethod
    def plot_individual_metric(metrics_df, exp_name=None, pca=True, offset_db=4):
        metrics = ["Silhoute_cosine", "Calinski_habrasz", "Davies_bouldin"]

        for metric in metrics:
            if pca:
                title = "{}. PCA".format(metric)
            else:
                title = "{}. No PCA".format(metric)

            ascending = False
            if metric == "Davies_bouldin":
                ascending = True
                title += " (lower the better)"
            else:
                title += " (higher the better)"

            if exp_name != None:
                title = "{} {}".format(title, exp_name)

            fig_1 = px.bar(metrics_df, y=metric, x="Cluster", title="{} {}".format("", title), color="cluster_type")
            fig_1.update_yaxes(title_text="{} Avg".format(metric))
            fig_1.update_xaxes(title_text="Experiment")

            if metric == "Davies_bouldin":
                fig_1.update(layout_yaxis_range=[0, offset_db])

            best_sill = metrics_df[~metrics_df["Exp"].str.contains("2|3")].sort_values(by=[metric], ascending=ascending)[:3]
            offset = best_sill[metric].mean() / 5

            i = 0
            for _, row in best_sill.iterrows():
                i += 1
                fig_1.add_annotation(x=row["Cluster"], y=row[metric] + offset, text="*" * i, showarrow=False)

            fig_1.show()

    def run_clusters(self, label="dmy", showFigs=False, mevsMut=None, clusters_config=None):
        if mevsMut is None:
            # the linter didn't like the nested conditions
            if hasattr(self, "mevsMut"):
                data = self.mevsMut
        else:
            data = mevsMut

        selected_clusters = ["Birch", "RawKMeans", "GaussianMixture", "Ward", "SpectralClustering", "Avg"]
        # selected_clusters = ["RawKMeans", "DBSCAN", "OPTICS", "SpectralClustering", "Avg"]

        # run experiments
        outputs, _, all_metrics, _ = cs.compare_exp(
            data,
            rob_comp=None,
            n_clusters=None,
            selected_clusters=selected_clusters,
            show_figures=False,
            show_consensus=True,
            pca_data=False,
            n_comp=5,
            default_base=clusters_config,
        )
        outputs.set_index("Sample", inplace=True)

        if showFigs:
            # Plot the metrics
            # fig = cs.display_metrics(all_metrics, "Cluster metrics for {}".format(exp.type), show_individual=False)
            # fig.show()
            all_metrics["cluster_type"] = ["-".join(cluster.split("_")[:1]) for cluster in all_metrics["Cluster"]]
            NetworkOutput.plot_individual_metric(all_metrics, pca=False, offset_db=4, exp_name=self.type)

        new_cols = [col + "_" + label for col in outputs.columns[2:]]
        outputs.columns = ["PC_1", "PC_2"] + new_cols

        return outputs


    # Helper
    @staticmethod
    def extract_tf_number(exp_name):
        # Regex to match a number followed by "TF"
        match = re.search(r'(\d+)TF', exp_name)
        if match:
            return int(match.group(1))  # Extract the number as an integer
        return None  # Return None if no match is found




#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   GraphHelper.py
@Time    :   2023/10/18 13:23:29
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   File that contains helper functions to analyse the different networks
"""

import numpy as np
import pandas as pd

# ploting libs=
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from sklearn.decomposition import PCA

from .utilities import clustering as cs
from .utilities import pre_processing as pre
from .utilities import sankey_consensus_plot as sky
from .GraphToolExp import GraphToolExperiment as GtExp
from .utilities.helpers import save_fig
from .ExperimentSet import ExperimentSet
from .NetworkOutput import NetworkOutput


########## Clustering analysis ##########
def plot_individual_metric(metrics_df, pca=True, export=False, offset_db=4, base_path=None):
    metrics = ["Silhoute_cosine", "Calinski_habrasz", "Davies_bouldin"]

    for metric in metrics:
        if pca:
            title = "{}. PCA".format(metric)
        else:
            title = "{}. No PCA".format(metric)

        ascending = False
        if metric == "Davies_bouldin":
            ascending = True
            title += " (lower the better)"
        else:
            title += " (higher the better)"

        fig_1 = px.bar(metrics_df, y=metric, x="Cluster", title="{} {}".format("", title), color="cluster_type")
        fig_1.update_yaxes(title_text="{} Avg".format(metric))
        fig_1.update_xaxes(title_text="Experiment")

        if metric == "Davies_bouldin":
            fig_1.update(layout_yaxis_range=[0, offset_db])

        best_sill = metrics_df[~metrics_df["Exp"].str.contains("2|3")].sort_values(by=[metric], ascending=ascending)[:3]

        offset = best_sill[metric].mean() / 5

        i = 0
        for _, row in best_sill.iterrows():
            i += 1
            fig_1.add_annotation(x=row["Cluster"], y=row[metric] + offset, text="*" * i, showarrow=False)

        if export:
            # save_fig(
            #     fig_1,
            #     title="{}".format(title),
            #     filename="/cluster_models/{}_best3".format(title),
            #     path=base_path,
            #     x_label="Experiment",
            #     y_label="{} Avg".format(metric),
            #     width=1920,
            #     height=800,
            # )
            save_fig(name="{}".format(title), fig=fig_1, width=1920, height=800, base_path=base_path)

        fig_1.show()


def run_clusters(exp: NetworkOutput, label="", show_figs=False, individual_plots=False, ge_exp=None, norm=False):
    if ge_exp is None:
        data = exp.mevsMut
    else:
        data = ge_exp

    if norm:
        data = data / data.max()

    selected_clusters = ["Birch", "RawKMeans", "GaussianMixture", "Ward", "SpectralClustering", "Avg"] 

    # run experiments
    outputs, _, all_metrics, _ = cs.compare_exp(
        data, rob_comp=None, n_clusters=None, selected_clusters=selected_clusters, show_figures=False, show_consensus=True, pca_data=False
    )
    outputs.set_index("Sample", inplace=True)

    fig = {}
    if show_figs:
        # Plot the metrics
        fig = cs.display_metrics(all_metrics, f"Cluster metrics for {exp.type}", show_individual=False, verbose=True)

        if individual_plots:
            plot_individual_metric(all_metrics, pca=False, offset_db=4)

    new_cols = [col + "_" + label for col in outputs.columns[2:]]
    outputs.columns = ["PC_1", "PC_2"] + new_cols

    return outputs, fig, all_metrics


def sill_distrib(metrics, label="Standard", points=None, figures_path=None):
    comb_df = pd.DataFrame()
    for _, row in metrics.iterrows():
        comb_df = pd.concat([comb_df, pd.DataFrame(row["Samples_Silhouete_cos"][1], columns=[row["Cluster"]])], axis=1)
    comb_df = comb_df.melt(var_name="Exp", value_name="Sillhouette (cosine)")
    comb_df = pd.concat([comb_df, (comb_df["Exp"].str.split("_", expand=True)[[0, 2]]).rename(columns={0: "Model", 2: "Cluster Size"})], axis=1)

    fig = px.box(comb_df, x="Cluster Size", y="Sillhouette (cosine)", color="Model", points=points, title=label)
    fig = fig.update_traces(boxmean=True)
    if figures_path is not None:
        fig = fig.update_layout(
            legend=dict(
                orientation="h",
                yanchor="bottom",
                xanchor="center",
                y=1.05,
                x=0.5,
                bgcolor="rgba(0,0,0,0)",
                font=dict(size=16, color="#003366"),
            ),
            font=dict(size=14),
            xaxis=dict(tickfont=dict(size=16)),
            yaxis=dict(tickfont=dict(size=16)),
            title="",
        )
        save_fig(name="{}_sill_spread".format(label), fig=fig, base_path=figures_path, width=1720, height=600)
    return fig, comb_df


def rank_cs_metrics(metrics: pd.DataFrame, label="std", figures_path=None, path=None):
    top_3 = pd.DataFrame()
    scores = ["Silhoute_cosine", "Calinski_habrasz", "Davies_bouldin"]
    for score in scores:
        ascending = False
        if score == "Davies_bouldin":
            ascending = True
        best_score = metrics.loc[~metrics["Exp"].str.contains("2|3")].sort_values(by=[score], ascending=ascending)[:3]
        r_score = score.split("_")[0][:3] + "_" + score.split("_")[1][:3]
        best_score["Score"] = r_score
        sel_df = best_score[["Cluster"]].transpose()
        sel_df.columns = ["{}_{}".format(r_score, idx) for idx in range(1, 4)]
        top_3 = pd.concat([top_3, sel_df], axis=1)

    abreviate_cs = {"Avg": "Avg", "RawKMeans": "KM", "Ward": "Wrd", "Birch": "Brch", "SpectralClustering": "Spec"}
    dmy_df = top_3.loc["Cluster"].str.split("_", expand=True)
    dmy_df[0].replace(abreviate_cs, inplace=True)
    abb_models = dmy_df[0] + "_" + dmy_df[2].astype(str)

    sel_df = pd.concat([top_3.transpose(), pd.Series(abb_models, name=label)], axis=1).drop(columns=["Cluster"])

    if path != None:
        filename = "{}.tsv".format(label)
        sel_df.reset_index(names="Metric_Rank").to_csv(figures_path + filename, sep="\t")

    return sel_df


def prcs_top3_metrics(comb_df: pd.DataFrame, label="top3", figures_path=None):
    # comb_df = pd.concat([top_std, top_rwd, top_pen], axis=1)
    # print(comb_df.columns)

    # melt the data
    tst_df = comb_df.melt(value_name="gen")
    tst_df = pd.concat([tst_df, tst_df["gen"].str.split("_", expand=True).rename(columns={0: "model", 1: "size"})], axis=1)
    tst_df["cat_gen"] = tst_df["gen"].astype("category").cat.codes
    tst_df["cat_model"] = tst_df["model"].astype("category").cat.codes
    tst_df["cat_size"] = tst_df["size"].astype("category").cat.codes

    figs = []
    for col in ["cat_gen", "cat_model", "cat_size"]:
        # update to numeric values
        coord_val = {row["gen"]: row[col] for _, row in tst_df.iterrows()}
        coord_val = comb_df.replace(coord_val)

        # compute the color
        # z_color = coord_val.replace(tst_df[col].value_counts().to_dict()).values
        z_color = []
        for net in tst_df["variable"].unique():
            sel = tst_df.loc[tst_df["variable"] == net]
            to_replace = sel[col].value_counts().to_dict()
            z_color.append(sel[col].replace(to_replace).values)

        z_color = np.array(z_color).T

        # the string value like the model, size etc
        orig_col = col.split("_")[-1]
        z_text = {row[col]: row[orig_col] for idx, row in tst_df.iterrows()}
        z_text = coord_val.replace(z_text)

        fig = px.imshow(coord_val, text_auto=True, aspect="auto", color_continuous_scale="sunset")
        fig = fig.update_traces(z=z_color, text=z_text, texttemplate="%{text}")

        fig = fig.update_layout(
                font=dict(size=18),
                xaxis=dict(tickfont=dict(size=18)),
                yaxis=dict(tickfont=dict(size=18)),
                title="",
        )

        if figures_path != None:
            save_fig(name="top3_cs_{}_{}".format(orig_col, label), fig=fig, base_path=figures_path, width=900, height=500, margin=0.02)

        figs.append(fig)

    if figures_path != None:
        comb_df.to_excel("{}/top3_cs_{}.xlsx".format(figures_path, label))

    return figs


def find_pcs(exp_df: pd.DataFrame):
    """

    The DataFrame needs to be in the form of n_samples x m_features

    Args:
        exp_df (pd.DataFrame): The data to which PCA is applied
    """

    for n_comp in range(2, exp_df.shape[1] - 1):
        pca_model = PCA(n_components=n_comp)
        pca_model.fit_transform(exp_df.values)

        if pca_model.explained_variance_ratio_.sum() >= 0.9:
            break

    print("Sum of 90% variance at PC: {}".format(n_comp))

    ## Number of components that there is a smaller change of 1%
    for n_comp in range(2, exp_df.shape[1] - 1):
        pca_model = PCA(n_components=n_comp)
        pca_model.fit_transform(exp_df.values)

        if np.where(np.diff(-pca_model.explained_variance_ratio_) <= 0.01)[0].size != 0:
            break

    print("Change < 1% at PC: {}".format(n_comp))


########## Analysis of gene representation ##########
def mev_modcon_genes(sel_exp: NetworkOutput, ref_tpms: pd.DataFrame, num_genes=100):
    sort_col = "ModCon_{}".format(sel_exp.type)
    sel_exp.mevsMut, meta = sel_exp.get_mevs(tpms=ref_tpms, modCon=sel_exp.modCons, sort_col=sort_col, num_genes=num_genes, verbose=True)

    data = []
    for key, val in meta.items():
        data.append((key, len(val["matched"]), len(val["not_matched"])))

    data_df = pd.DataFrame(data, columns=["Comm", "Matched", "Not matched"])
    data_df = data_df.melt(id_vars="Comm", value_vars=["Matched", "Not matched"], var_name="Type", value_name="Num genes")
    data_df["Comm"] = data_df["Comm"].astype(str)

    return data_df


def scatter_network_stats(df, metric="degree", height=700, width=1200):
    degree_cols = [col for col in df.columns if metric in col]

    # prepare dataframe for plotting
    # 1. Metl the dataframes
    dmy_df = pd.melt(df[degree_cols], ignore_index=False, value_name=metric, var_name="TF_metric")
    dmy_df["is_tf"] = df["is_tf"]
    # 2. extract the number of TF
    dmy_df["TF_num"] = [int(elem.split("_")[1].split("TF")[0]) for elem in dmy_df["TF_metric"]]
    dmy_df.sort_values(by=["TF_num", "is_tf"], ascending=True, inplace=True)
    dmy_df.reset_index(names="gene", inplace=True)

    fig = px.scatter(
        dmy_df,
        x="gene",
        y=metric,
        color="is_tf",
        # color_discrete_sequence=["red", "blue"],
        category_orders={"is_TF": ["1", "0"]},
        facet_col="TF_num",
        title="Healthy gc42, standard 5K. Metric: {}".format(metric),
        facet_col_wrap=4,
        height=height,
        width=width,
    )
    return fig, dmy_df


def get_network_metrics(exp_set: ExperimentSet, selected_exps=None):
    if selected_exps is None:
        selected_exps = exp_set.get_exp_labels()

    # Select experiments
    all_nodes_df = pd.DataFrame()
    for exp in exp_set.get_exps():
        if exp.type not in selected_exps:
            continue

        # proces the network stats graph
        g_stats = exp.compute_graph_stats()
        tf_type = exp.name.split("_")[-1]
        remap_cols = {col: f"{col}_{tf_type}" for col in g_stats.columns}
        g_stats.rename(columns=remap_cols, inplace=True)

        # proces the nodes_df
        dmy = exp.nodes_df.copy(deep=True)
        # dmy.rename(columns={"Modularity Class": "Leiden"}, inplace=True)
        used_cols = ["TF", "IVI"]
        dmy = dmy[used_cols]
        remap_cols = {col: f"{col}_{tf_type}" for col in used_cols}
        dmy.rename(columns=remap_cols, inplace=True)

        # combine them
        dmy_comb = pd.concat([g_stats, dmy], axis=1)
        # dmy_comb["tf_num"] = tf_type.split("TF")[0]

        all_nodes_df = pd.concat([all_nodes_df, dmy_comb], axis=1)

    # add marker for a gene if is TF
    all_nodes_df = pd.concat([all_nodes_df, exp.nodes_df["TF"].astype(str)], axis=1)
    # all_nodes_df["is_tf"] = exp.nodes_df["TF"].astype(str)
    all_nodes_df = all_nodes_df.rename(columns={"TF": "is_tf"})
    return all_nodes_df


def prep_net_metrics(std_nt: NetworkOutput, rwrd_nt: NetworkOutput, pen_nt: NetworkOutput):
    std_nt.export_to_gephi()
    rwrd_nt.export_to_gephi()
    pen_nt.export_to_gephi()

    std_sts = std_nt.compute_graph_stats()
    rwrd_sts = rwrd_nt.compute_graph_stats()
    pen_sts = pen_nt.compute_graph_stats()

    std_sts["IVI"] = std_nt.nodes_df["IVI"]
    rwrd_sts["IVI"] = rwrd_nt.nodes_df["IVI"]
    pen_sts["IVI"] = pen_nt.nodes_df["IVI"]

    std_sts["Type"] = "Standard"
    rwrd_sts["Type"] = "Reward"
    pen_sts["Type"] = "Penalised"

    combined_df = pd.concat([std_sts, rwrd_sts, pen_sts], axis=0).reset_index(names="gene")

    combined_df["TF"] = combined_df["gene"].apply(lambda gene: std_nt.nodes_df.loc[gene]["TF"])
    combined_df["TF"].fillna(0, inplace=True)
    combined_df["TF"] = combined_df["TF"].astype(str)

    return combined_df


def prep_single_net_metrics(network: NetworkOutput, label="Standard"):
    network.export_to_gephi()

    sts = network.compute_graph_stats()
    sts["IVI"] = network.nodes_df["IVI"]

    sts.reset_index(names="gene", inplace=True)
    sts["TF"] = sts["gene"].apply(lambda gene: network.nodes_df.loc[gene]["TF"])
    sts["TF"].fillna(0, inplace=True)
    sts["TF"] = sts["TF"].astype(str)
    sts["Type"] = label

    return sts


def plot_net_metrics(metrics_df, label, log_y=False, color="Type", figs_path=None, filename="NetMetrics"):
    metrics = ["degree", "pageRank", "closeness", "betwenees", "IVI"]

    figs, titles = [], []
    for metric in metrics:
        figs.append(
            px.box(metrics_df, y=metric, x="Type", color=color, points="all", log_y=log_y, hover_data=["gene", "Type", "TF", metric],
                            # category_orders={'Type': ['Tum Std', 'Non-Tum Std', 'Tum Rwd', 'Non-Tum Rwd']}
                    ) 
        )
        titles.append(metric)

    subplots_config = {
        "num_cols": 2,
        "shared_x": True,
        "shared_y": False,
        "h_spacing": None,
        "v_spacing": 0.05,
        "main_title": "Different network metrics {}".format(label),
        "height": 700,
        "width": None,
        "y_title": None,
        "x_title": None,
        "specs": [[{}, {}], [{}, {}], [{"colspan": 2}, None]],
    }

    fig = helper_multiplots(figs, titles, subplots_config)
    if log_y:
        fig = fig.update_yaxes(type="log")

    if figs_path:
        fig.update_layout(
            title="",
            xaxis=dict(
                tickfont=dict(size=20),
            ),
            yaxis=dict(
                tickfont=dict(size=20),
            ),
            font=dict(size=18),
            template='ggplot2'
            # paper_bgcolor="rgba(0,0,0,0)",
        )
        fig.update_traces(marker_size=8)
        save_fig(name=filename, fig=fig, base_path=figs_path, width=1400, height=1000, margin=0.02)

    return fig


########## Overview of Sankey/Leiden plots ##########
def plot_sankey_leiden(exps: ExperimentSet, sky_fig: dict, rename_cols: str, tf="6", label=""):
    _, all_leiden = exps.avg_leiden_scores()

    # filtering an pre-processing
    all_leiden = all_leiden.loc[all_leiden["TF"] == tf]
    all_leiden.loc[all_leiden["Modifier"] == "beta", "Modifier"] = "Penalised"

    figs, titles = [], []
    for metric in ["ModuleNum", "ModularityScore"]:
        fig = px.box(all_leiden, x="Modifier", y=metric, color="TF", points="all", category_orders={"Modifier": ['Standard', "Reward", "Penalised"]})
        figs.append(fig)

    figs.append(sky_fig)

    titles.append("Communities Number")
    titles.append("Modularity Score")
    titles.append("")  # None for Sankey

    subplots_config = {
        "num_cols": 2,
        "shared_x": True,
        "shared_y": False,
        "h_spacing": None,
        "v_spacing": 0.1,
        "main_title": "Modularity scores vs community size {}".format(label),
        "height": 700,
        "width": None,
        "y_title": None,
        "x_title": None,
        "specs": [[{}, {}], [{"colspan": 2, "type": "sankey"}, None]],
    }

    fig = helper_multiplots(figs, titles, subplots_config)

    start_x = 0.00
    step = 1 / (len(rename_cols.values()) - 1)
    for idx, col_name in enumerate(rename_cols.values()):
        x_pos = round(start_x + (idx * step), 2)
        fig.add_annotation(x=x_pos, y=0.47, xref="paper", yref="paper", text=col_name, showarrow=False, font=dict(size=20, color="black"))

    return fig


def prep_sankey_leiden(exps: ExperimentSet, vu_output: pd.DataFrame, tf=6, no_genes="4K", no_K=6, chosen_cs_model="RawKMeans"):
    comb_std, _, _ = run_clusters(exps.exps["standard_{}_{}TF".format(no_genes, tf)], label="std_tf{}".format(tf))
    comb_norm3, _, _ = run_clusters(exps.exps["norm3_{}_{}TF".format(no_genes, tf)], label="norm3_tf{}".format(tf), show_figs=False)
    comb_norm3.drop(columns=["PC_1", "PC_2"], inplace=True)
    comb_beta, _, _ = run_clusters(exps.exps["beta_{}_{}TF".format(no_genes, tf)], label="beta_tf{}".format(tf))
    comb_beta.drop(columns=["PC_1", "PC_2"], inplace=True)

    sel_exp = exps.exps["standard_4K_{}TF".format(tf)]
    comb_df = pd.concat([comb_std, comb_norm3, comb_beta, vu_output], axis=1).dropna()

    reorder_cols = [
        "TCGA408_classifier",
        "KMeans_labels_6",
        "{}_CS_{}_std_tf{}".format(chosen_cs_model, no_K, tf),
        "{}_CS_{}_norm3_tf{}".format(chosen_cs_model, no_K, tf),
        "{}_CS_{}_beta_tf{}".format(chosen_cs_model, no_K, tf),
        "2019_consensus_classifier",
    ]

    rename_cols = {
        reorder_cols[0]: "TCGA",
        reorder_cols[1]: "CA + IFNg",
        reorder_cols[2]: "Standard",
        reorder_cols[3]: "Reward",
        reorder_cols[4]: "Penalised",
        reorder_cols[5]: "Consensus",
    }

    _, sky_fig = sky.main(
        df=comb_df.rename(columns=rename_cols),
        reorder_cols=list(rename_cols.values()),
        title="Best for {}. Comp between {} ".format(sel_exp.type, ", ".join(reorder_cols)),
        retMeta=True,
    )

    return exps, sky_fig, rename_cols


def gene_stats(df: pd.DataFrame, tf_list: pd.DataFrame, mut_df: pd.DataFrame, mut_count=1):
    df["mut_count"] = mut_df["count"]
    df.fillna(0, inplace=True)

    # Compute stats
    stats = {}

    stats_idx = ["Mut", "Mut_TF"]  # "TF"]
    data = []
    for stat in stats_idx:
        df_key = f"{stat}_df"
        if stat == "Mut":
            stats[df_key] = df.loc[df["mut_count"] >= mut_count]
            # stats[f"{stat}_prct"] = stats[df_key].shape[0] / df.shape[0] * 100
        # elif stat == "TF":
        #     stats[df_key] = df.loc[df.index.isin(tf_list)]
        elif stat == "Mut_TF":
            included_tf = df[df.index.isin(tf_list)]
            stats[df_key] = included_tf.loc[included_tf["mut_count"] >= mut_count]
            # stats[f"{stat}_prct"] = stats[df_key].shape[0] / included_tf.shape[0] * 100

        stats[f"{stat}_prct"] = stats[df_key].shape[0] / df.shape[0] * 100
        stats[f"{stat}_prct"] = round(stats[f"{stat}_prct"], 2)
        stats[f"{stat}_num"] = stats[df_key].shape[0]
        data.append([stats["{}_num".format(stat)], stats["{}_prct".format(stat)]])

    stats_df = pd.DataFrame(
        data,
        columns=["Num", "Prct"],
        index=stats_idx,
    )

    stats_df["Burden"] = ">={}".format(mut_count)

    return stats, stats_df


def stats_mut_burden(exp_df: pd.DataFrame, tf_list: pd.DateOffset, mut_df: pd.DataFrame, type="a_type"):
    comb_df = pd.DataFrame()
    for mut_count in range(0, 11, 1):
        _, dmy_df = gene_stats(exp_df, tf_list, mut_df, mut_count=mut_count)

        comb_df = pd.concat([comb_df, dmy_df], axis=0)

    comb_df["Type"] = type

    return comb_df


########## Difference in gene selection ##########
def extract_gene_sel(exp: NetworkOutput, ref_ge: pd.DataFrame, num_genes=3000, verbose=True):
    """
    Find the differences in genes between a network and highest std/med variance

    Args:
        exp (PGCNAOutput): The experiments were to perform
        ref_ge (pd.DataFrame): Gene expression reference
        num_genes (int, optional): Number of genes selected for the highest std/median. Defaults to 3000.

    Returns:
        stats (dict): Useful stats
    """
    # get the Cluster analysis genes
    cs_genes = pre.select_genes(ref_ge.reset_index(), no_genes=num_genes, relative_selection=True)

    # generate MEVs for the given experiment
    modCon = exp.get_ModCon()
    if (type(exp) == GtExp) or (hasattr(exp, 'sbm_method')):  # check doesn't work for the first experiments with sbm
        sort_col = "ModCon_{}_gt".format(exp.type)
    else:
        sort_col = "ModCon_{}".format(exp.type)

    # sort_col = "ModCon_{}_gt".format(exp.type)

    exp.mevsMut, meta = exp.get_mevs(tpms=ref_ge, modCon=modCon, sort_col=sort_col, num_genes=100, verbose=True)

    net_sel_genes, net_all_genes = [], []
    for key, val in modCon.items():
        net_sel_genes.extend(meta[key]["modCon_genes"])
        net_all_genes.extend(val.index)

    stats = {
        "diff_all": set(cs_genes) - set(net_all_genes),
        "diff_sel": set(cs_genes) - set(net_sel_genes),
        "cmn_sel": set(cs_genes) & set(net_sel_genes),
        "cmn_all": set(cs_genes) & set(net_all_genes),
        "cs_genes": cs_genes,
        "net_sel_genes": net_sel_genes,
    }

    if verbose:
        print(f"CS vs Network (all). There are {len(stats['diff_all'])} different genes. ")
        print(f"CS vs Network (sel). There are {len(stats['diff_sel'])} different genes. ")
        print(f"CS vs Network (sel). Common genes: {len(stats['cmn_sel'])}")
        print(f"CS vs Network (all). Common genes: {len(stats['cmn_all'])}")
        print(f"Network selected: {len(stats['net_sel_genes'])}")
        print(f"Highest relative/std {len(stats['cs_genes'])}")

    return stats


def prep_net_vs_ca(ge_df: pd.DataFrame, mut_df: pd.DataFrame, gt_genes, cs_genes):
    """
    Function that prepares the given Gene Expression DataFrame to mark the genes included by both Network and Clustering Analysis method

    Args:
        ge_df (pd.DataFrame): All the GE dataframe
        mut_df (pd.DataFrame): The TCGA mut df for the mutation count
        gt_genes (_type_): The genes selected by the network approach
        cs_genes (_type_): The genes selected by the cluster analysis approach

    Returns:
        dmy_df (DataFrame): Procssed dataframe
    """
    dmy_df = ge_df[ge_df.index.isin(list(set(cs_genes + gt_genes)))].copy(deep=True)
    median_raw = dmy_df.median(axis=1)
    dmy_df = np.log2(dmy_df + 1)
    std, median = dmy_df.std(axis=1), dmy_df.median(axis=1)
    dmy_df["rel_var"] = std / median
    dmy_df["std"] = std
    dmy_df["median_log"] = median
    dmy_df["median_raw"] = median_raw

    dmy_df["type"] = "Both"
    dmy_df.loc[dmy_df.index.isin(list(set(cs_genes) - set(gt_genes))), "type"] = "CS"
    dmy_df.loc[dmy_df.index.isin(list(set(gt_genes) - set(cs_genes))), "type"] = "Network"

    dmy_df["mut_count"] = mut_df["count"]
    dmy_df.fillna(0, inplace=True)
    dmy_df["mut_count"] = dmy_df["mut_count"] + 1

    return dmy_df


def plot_net_vs_ca(prep_df: pd.DataFrame, log=False, annotations=False, ann_add=None, ann_rm=None, ann_chg=None):
    """
    Plotting the gene selection between Network and Clustering anlasysis

    Args:
        prep_df (pd.DataFrame): The DataFrame after prep func
        log (bool, optional): Axes in log scale. Defaults to False.
        annotations (bool, optional): Add annotations. Defaults to False.
        ann_add (list, optional): Annotate the given list of genes. Defaults to [].
        ann_rm (list, optional): Remove the annotations for the given list. This is used for declutering. Defaults to [].
        ann_chg (list, optional): Change the annotation offset for a given list of genes. Defaults to [].

    Returns:
        Plotly figure: Ploly Figure
    """
    fig = px.scatter(
        prep_df.reset_index(),
        y="mut_count",
        x="rel_var",
        color="type",
        title="Network vs Clustering Analysis",
        size="median_raw",
        size_max=60,
        height=700,
        hover_data=["genes", "mut_count", "rel_var"],
        log_y=log,
        log_x=log,
        # marginal_x="box",
        # marginal_y="box",
    )

    fig = fig.update_yaxes(title="Mut count")
    fig = fig.update_xaxes(title="Relative variance log2(TPM+1)")

    if annotations:
        marker_genes, num_genes = [], 20

        # prep_df = prep_df.loc[prep_df.index.values != "TP53"]
        marker_genes.extend(prep_df.sort_values(by=["rel_var"], ascending=False).index.values[:num_genes])
        marker_genes.extend(prep_df.sort_values(by=["mut_count"], ascending=False).index.values[:num_genes])

        if len(ann_add) > 0:
            marker_genes.extend(ann_add)
        if len(ann_rm) > 0:
            marker_genes = set(marker_genes) - set(ann_rm)

        counter = 0
        offset = 15
        for idx, row in prep_df.loc[prep_df.index.isin(marker_genes)][["rel_var", "mut_count"]].iterrows():
            yshift = offset
            if counter % 2 == 0:
                yshift = -offset

            if idx in ann_chg:
                yshift = offset
            elif idx == "":
                yshift = -offset

            if log:
                fig.add_annotation(
                    x=np.log10(row["rel_var"]), y=np.log10(row["mut_count"]), text=idx, showarrow=False, valign="bottom", xshift=yshift, yshift=yshift
                )
            else:
                fig.add_annotation(x=row["rel_var"], y=row["mut_count"], text=idx, showarrow=False, valign="bottom", xshift=yshift, yshift=yshift)

    return fig


##### Helper functions - These should be in a seperate file/class #####
def plot_leiden(exps: ExperimentSet, tf="6", label=""):
    # filtering an pre-processing
    _, all_leiden = exps.avg_leiden_scores()

    all_leiden = all_leiden.loc[all_leiden["TF"] == tf]
    all_leiden.loc[all_leiden["Modifier"] == "beta", "Modifier"] = "Penalty"

    figs, titles = [], []
    for metric in ["ModuleNum", "ModularityScore"]:
        fig = px.box(all_leiden, x="Modifier", y=metric, color="TF", points="all")
        figs.append(fig)

    titles.append("Communities Number")
    titles.append("Modularity Score")

    subplots_config = {
        "num_cols": 2,
        "shared_x": True,
        "shared_y": False,
        "h_spacing": None,
        "v_spacing": 0.05,
        "main_title": f"Modularity scores vs community size TF-{tf} {label}",
        "height": 500,
        "width": None,
        "y_title": None,
        "x_title": None,
        "specs": None,
    }

    fig = helper_multiplots(figs, titles, subplots_config)
    return fig


########## General plotting functions ##########


def plot_metrics_distrib(df, metric="degree", exp_label="Experiment", tf_range=None, x_range=None, path=None, filename="MetricsDistrib"):
    # metric = "degree"
    # df = all_nodes_df.copy(deep=True)
    if tf_range is None:
        tf_range = range(3, 15)

    metric_cols = [col for col in df.columns if metric in col]

    # prepare dataframe for plotting
    # 1. Melt the dataframes
    dmy_df = pd.melt(df[metric_cols], ignore_index=False, value_name=metric, var_name="TF_metric")
    dmy_df["is_tf"] = df["is_tf"]
    # 2. extract the number of TF and limit to 15
    dmy_df["TF_num"] = [int(elem.split("_")[1].split("TF")[0]) for elem in dmy_df["TF_metric"]]
    dmy_df.sort_values(by=["TF_num", "is_tf"], ascending=True, inplace=True)
    dmy_df.reset_index(names="gene", inplace=True)
    dmy_df = dmy_df[dmy_df["TF_num"].isin(tf_range)]

    figs, titles = [], []
    for num_tf in dmy_df["TF_num"].unique():
        is_tf = dmy_df.loc[(dmy_df["TF_num"] == num_tf) & (dmy_df["is_tf"] == "1")][metric]
        non_tf = dmy_df.loc[(dmy_df["TF_num"] == num_tf) & (dmy_df["is_tf"] == "0")][metric]
        hist_data = [is_tf, non_tf]
        group_labels = ["is_tf", "non_tf"]

        fig = ff.create_distplot(hist_data, group_labels, show_rug=False)
        # fig = fig.update_layout(xaxis=dict(tickmode="linear", tick0=3, dtick=5))
        if x_range is not None:
            fig = fig.update_layout(xaxis=dict(range=x_range))
        figs.append(fig)
        titles.append("plot_net_metricsTF_{}".format(num_tf))

    subplots_config = {
        "num_cols": 4,
        "shared_x": False,
        "shared_y": True,
        "h_spacing": 0.03,
        "v_spacing": 0.1,
        "main_title": "{} for {}. KDE distribution.".format(metric, exp_label),
        "height": 700,
        "width": None,
        "x_title": metric,
        "y_title": "Probability",
        "specs": None,
    }

    fig = helper_multiplots(figs, titles, subplots_config)

    # zoom in if specified
    if x_range is not None:
        fig = fig.update_xaxes(range=x_range)

    if path:
        save_fig(name=filename, fig=fig, base_path=path, width=1700, height=900)

    return fig


def plot_mut_rep(df: pd.DataFrame, title: str):
    fig = px.bar(
        df.reset_index(names="Gene_Type"),
        x="Burden",
        # y="Prct",
        y='Num',
        color="Type",
        barmode="group",
        facet_row="Gene_Type",
        title=title,
        text_auto=True,
        height=700,
    )

    fig = fig.update_layout(font=dict(size=16))
    return fig


def helper_multiplots(figs, subplot_tiles, config=None):
    if config is None:
        config = {
            "num_cols": 2,
            "shared_x": False,
            "shared_y": True,
            "h_spacing": 0.05,
            "v_spacing": 0.15,
            "main_title": "Subplots",
            "height": 700,
            "width": None,
            "y_title": "Y-axis",
            "x_title": "X-axis",
            "specs": None,
        }

    num_cols = config["num_cols"]
    num_rows = int(np.ceil(len(figs) / num_cols))

    subplot = make_subplots(
        rows=num_rows,
        cols=num_cols,
        subplot_titles=subplot_tiles,
        shared_xaxes=config["shared_x"],
        shared_yaxes=config["shared_y"],
        horizontal_spacing=config["h_spacing"],
        vertical_spacing=config["v_spacing"],
        specs=config["specs"],
    )

    idx_row, idx_col = 1, 1
    for i, fig in enumerate(figs):
        for trace in range(len(fig["data"])):
            subplot.add_trace(fig["data"][trace], row=idx_row, col=idx_col)

        if idx_col % num_cols == 0:
            idx_col = 0
            idx_row += 1
        idx_col += 1

    layout = go.Layout(title_text=config["main_title"])

    subplot = subplot.update_layout(layout, height=config["height"], width=config["width"], showlegend=False, title_y=0.9)
    subplot = subplot.update_xaxes(title_text=config["x_title"])
    subplot = subplot.update_yaxes(title_text=config["y_title"])

    # Remove duplicate legend items
    # visited = []
    # for trace in subplot["data"]:
    #     if trace["name"] is None:
    #         continue

    #     trace["name"] = trace["name"].split(",")[0]
    #     if trace["name"] not in visited:
    #         trace["showlegend"] = True
    #         visited.append(trace["name"])
    #     else:
    #         trace["showlegend"] = False

    return subplot


def helper_multiplots_hist(figs, subplot_titles, df: pd.DataFrame, config=None, color_cols=None):

    colors = px.colors.qualitative.G10 + px.colors.qualitative.D3 + px.colors.qualitative.Alphabet

    if config is None:
        config = {
            "num_cols": 3,
            "shared_x": False,
            "shared_y": False,
            "h_spacing": 0.04,
            "v_spacing": 0.12,
            "main_title": "Metadata exploration",
            "height": 1000,
            "width": None,
            "y_title": None,
            "x_title": None,
            "specs": None,
        }

    num_cols = config["num_cols"]
    num_rows = int(np.ceil(len(figs) / num_cols))

    # Assuming `dmy_df` contains the column 'dendrogram_label'
    color_map = {}
    if color_cols != None:
        uniq_values = []
        for col in color_cols:
            uniq_values.extend(df[col].unique())
        uniq_values = list(set(uniq_values))

        # Create a dictionary to map each label to a color
        color_map = {label: colors[i % len(colors)] for i, label in enumerate(uniq_values)}

    subplot = make_subplots(
        rows=num_rows,
        cols=num_cols,
        subplot_titles=subplot_titles,
        shared_xaxes=config["shared_x"],
        shared_yaxes=config["shared_y"],
        horizontal_spacing=config["h_spacing"],
        vertical_spacing=config["v_spacing"],
        specs=config["specs"],
    )

    idx_row, idx_col = 1, 1
    visited = set()
    for idx, fig in enumerate(figs):
        for trace in fig.data:
            # Modify the trace directly
            trace_color = color_map[trace.name.split(": ")[1]] if ":" in trace.name else color_map[trace.name]

            showlegend = False
            if trace.name not in visited:
                showlegend = True
                visited.add(trace.name)

            modified_trace = go.Histogram(
                x=trace.x,
                name=trace.name,
                legendgroup=f"group_{idx}",  # Assign a unique legend group
                marker=dict(color=trace_color),
                showlegend=showlegend,
                texttemplate="%{y}",  # Display the count of each bar
            )
            subplot.add_trace(modified_trace, row=idx_row, col=idx_col)

        if idx_col == num_cols:
            idx_col = 1
            idx_row += 1
        else:
            idx_col += 1

    subplot.update_layout(height=config["height"], title_text=config["main_title"], showlegend=True)
    return subplot


def update_legend(fig, pos_x=0.95):
    return fig.update_layout(
        height=1200,
        title_y=0.97,
        title_x=0.5,
        legend=dict(
            title="",
            orientation="h",
            yanchor="top",
            y=1.0,
            xanchor="right",
            x=pos_x,
            bgcolor="rgba(0,0,0,0)",
        ),
    )


########## Helpers##########


def remap_columns(output_df):
    """
    Remaps the column names of the given DataFrame based on a predefined mapping dictionary.

    Args:
        output_df (pd.DataFrame): The DataFrame whose column names need to be remapped.

    Returns:
        pd.DataFrame: The DataFrame with remapped column names.
    """
    abreviate_cs = {"Avg": "Avg", "RawKMeans": "KM", "Ward": "Wrd", "Birch": "Brch", "SpectralClustering": "Spec", "GaussianMixture": "GMM"}
    remap_cols = {}
    for col in output_df.columns:
        dmy = col.split("_")[0]
        if dmy in abreviate_cs:
            remap_cols[col] = col.replace(dmy, abreviate_cs[dmy])
        else:
            remap_cols[col] = col

    return output_df.rename(columns=remap_cols)


def color_scale(color_scale="Sunset", num_points=100):

    import re  # Regular expression library for extracting RGB values

    # Assuming colors is a list of strings like 'rgb(75, 41, 145)'
    colors = px.colors.sample_colorscale(color_scale, samplepoints=num_points, colortype="rgb")

    # Function to convert an RGB string 'rgb(x, y, z)' to Hex
    def rgb_string_to_hex(rgb_string):
        # Extract the numerical values using regular expression
        numbers = re.findall(r"\d+", rgb_string)
        # Convert the string numbers to integers
        rgb = [int(num) for num in numbers]
        # Convert to hexadecimal
        return "#{0:02x}{1:02x}{2:02x}".format(rgb[0], rgb[1], rgb[2])

    # Convert each RGB string to Hex
    hex_colors = [rgb_string_to_hex(color) for color in colors]

    return hex_colors


def add_stats_to(nodes_df: pd.DataFrame, tpm_df: pd.DataFrame):

    nodes_df["mean"] = tpm_df.loc[nodes_df.index].mean(axis=1).round(4)
    nodes_df["median"] = tpm_df.loc[nodes_df.index].median(axis=1).round(4)
    nodes_df["std"] = tpm_df.loc[nodes_df.index].std(axis=1).round(4)
    nodes_df["var"] = tpm_df.loc[nodes_df.index].var(axis=1).round(4)
    return nodes_df


def export_top_modCon_genes(nodes_df: pd.DataFrame, path: str, top_n=50, metric="median"):
    comm_genes = {}
    ascending = metric == "ModCon_Rank"
    for com in nodes_df["max_b"].unique():
        genes = nodes_df[nodes_df["max_b"] == com].sort_values(by=metric, ascending=ascending).index

        comm_genes[com] = list(genes)[:50]
        # print(f"markers_{com} = {genes}")
        # print(f"\n\n custom_traces.append({{'genes': markers_{com}, 'title': 'Com_{com}'}})")

    comm_genes = pd.DataFrame().from_dict(comm_genes, orient="index").T
    comm_genes.to_csv(f"{path}/Top_{top_n}_{metric}.tsv", sep="\t", index=False)

    return comm_genes

########## Check functions ##########


def check_ctrls(controls: dict):
    ref_tf_ctrls = {}
    for key, ctrl in controls.items():
        same_across_ctrl, ref_tf = True, {}
        for tf, exp in ctrl["exps"].items():
            if len(ref_tf) == 0:
                ret_tf = set(exp.tf_list)
            else:
                if len(set(exp.tf_list) & set(ret_tf)) > 0:
                    print("TF ", tf)
                    same_across_ctrl = False
                    break

        ref_tf_ctrls[key] = ret_tf
        if not same_across_ctrl:
            print("Different TF list across Ctrl {}".format(key))

    diff_across_ctrl = True
    for key_1, val_1 in ref_tf_ctrls.items():
        ref_tf = val_1
        for key_2, val_2 in ref_tf_ctrls.items():
            if key_1 == key_2:
                continue

            # if they have all the genes common; We expect that there will be some random genes shared across the contro
            if len(val_1 & val_2) == len(ret_tf):
                print("Ctrl {} & Ctrl {} are the same".format(key_1, key_2))
                diff_across_ctrl = False
                break

        if not diff_across_ctrl:
            break

    if diff_across_ctrl:
        print("Different TF list across Controls ")



#!/usr/bin/env python
# -*-coding:utf-8 -*-
'''
@File    :   utilities.py
@Time    :   2023/06/08 08:55:34
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   This is a strip down version of utilities from Subtypes paper.
'''
import numpy as np
import pandas as pd
import plotly.express as px
from lifelines import KaplanMeierFitter


def save_fig(name, fig, width=None, height=None, scale=None, base_path=None, margin=None):
    """ Saves a figure to the base path or othwerise to th specified path

    Args:
        name ([type]): Name of the figure
        fig ([type]): Figure object
        width ([type], optional): Default 1280.
        height ([type], optional): Default 720.
        scale ([type], optional): Default 2.
        base_path ([type], optional): Default root folder.
    """

    default_width, default_height, default_scale = 1280, 720, 2
    default_base_path = "./data/figures"

    # override arguments if needed
    if width != None:
        default_width = width
    if height != None:
        default_height = height
    if scale != None:
        default_scale = scale
    if base_path != None:
        default_base_path = base_path

    if margin != None:
        margin_x = width * margin
        margin_y = height * margin

        fig.update_layout(
            margin=dict(l=margin_x, r=margin_x, t=margin_y+10, b=0)
        )

    fig.write_image(default_base_path + name + ".png",width=default_width, height=default_height, scale=default_scale)


def survival_plot(df, df_meta, classifier, selected_labels=None, color_map=None):

    dmy_meta = df_meta.loc[df_meta.index.isin(df.index.values)][["days_to_last_follow_up", "days_to_death"]]
    df = pd.concat([df, dmy_meta], axis=1)

    labels = list(df[classifier].unique())
    if selected_labels:
        labels = selected_labels

    models = []
    all_df = []
    for label in labels:
        kmf = KaplanMeierFitter()

        # process the data

        # older version for non-pyarrow format
        # dmy = df[df[classifier] == label][["days_to_last_follow_up", "days_to_death"]].replace("--", 0).astype(int)

        # to make it compatabile with pyarrow
        dmy = df[df[classifier] == label][["days_to_last_follow_up", "days_to_death"]].astype(str).replace("--", pd.NA).fillna(0).astype(float)

        dmy["last_contact"] = dmy[["days_to_last_follow_up", "days_to_death"]].max(axis=1)
        dmy["dead"] = np.where(dmy["days_to_death"] > 0, True, False)

        kmf.fit(dmy["last_contact"], event_observed=dmy["dead"], label=[label])
        models.append(kmf)

        # prepare df for plotting and change the scale from days to month
        dmy_df = kmf.survival_function_.copy(deep=True)

        disease = [dmy_df.columns[0][0]] * dmy_df.shape[0]
        dmy_df.rename(columns={dmy_df.columns[0][0]: "chance"}, inplace=True) 
        dmy_df.reset_index(inplace=True)
        chance = [value[0] for value in dmy_df["chance"].astype(float).values] 
        timeline = [value[0] /30 for value in dmy_df["timeline"].astype(int).values]

        dmy_df = pd.DataFrame()
        dmy_df["disease"] = disease
        dmy_df["timeline"] = timeline
        dmy_df["chance"] = chance
        all_df.append(dmy_df)

    test = pd.concat(all_df[:]).reset_index(drop=True)
    test["disease"] = test["disease"].astype(str)

    fig = px.line(test, x="timeline", y="chance", color="disease", markers=True, line_shape="hv", color_discrete_map = color_map) 
    fig.update_yaxes(title_text="Survival rate")
    fig.update_xaxes(title_text="Time (months)")
    fig.update_xaxes(range=[-1, 60])
    return fig

def survival_comp(df, df_meta, classifier_1, classifier_2, selected_labels_1 = None, selected_labels_2 = None,  color_map=None):

    # survival_comp(tum_h_tf10_cs.drop(columns=["days_to_last_follow_up", "days_to_death"]),  vu_output, classifier_1 = model, classifier_2="KMeans_labels_6", color_map=None)

    dmy_meta = df_meta.loc[df_meta.index.isin(df.index)][["days_to_last_follow_up", "days_to_death"]]
    df = pd.concat([df, dmy_meta], axis=1)
    labels = list(df[classifier_1].unique())
    if selected_labels_1:
        labels = selected_labels_1

    if selected_labels_2:
        labels.extend(selected_labels_2)
    else:
        labels.extend(list(df[classifier_2].unique()))

    labels = list(set(labels)) #eliminate duplicates

    models = []
    all_df = []
    for label in labels:
        kmf = KaplanMeierFitter()

        dmy = pd.DataFrame()
        # process the data
        if df[df[classifier_1] == label].shape[0]: 
            dmy = df[df[classifier_1] == label][["days_to_last_follow_up", "days_to_death"]] 

        if df[df[classifier_2] == label].shape[0]:
            dmy = df[df[classifier_2] == label][["days_to_last_follow_up", "days_to_death"]]

        # older version
        # dmy = dmy.replace("--", 0).astype(int)

        # to make it compatabile with pyarrow
        dmy = dmy.replace("--", pd.NA).fillna(0).astype(float)

        # dmy_1 = df[df[classifier_2] == label][["days_to_last_follow_up", "days_to_death"]].replace("--", 0).astype(int)

        dmy["last_contact"] = dmy[["days_to_last_follow_up", "days_to_death"]].max(axis=1)
        dmy["dead"] = np.where(dmy["days_to_death"] > 0, True, False)

        kmf.fit(dmy["last_contact"], event_observed=dmy["dead"], label=[label])
        models.append(kmf)

        # prepare df for plotting and change the scale from days to month
        dmy_df = kmf.survival_function_.copy(deep=True)

        disease = [dmy_df.columns[0][0]] * dmy_df.shape[0]
        dmy_df.rename(columns={dmy_df.columns[0][0]: "chance"}, inplace=True) 
        dmy_df.reset_index(inplace=True)
        chance = [value[0] for value in dmy_df["chance"].astype(float).values] 
        timeline = [value[0] /30 for value in dmy_df["timeline"].astype(int).values]

        dmy_df = pd.DataFrame()
        dmy_df["disease"] = disease
        dmy_df["timeline"] = timeline
        dmy_df["chance"] = chance
        all_df.append(dmy_df)

    test = pd.concat(all_df[:]).reset_index(drop=True)
    test["disease"] = test["disease"].astype(str)

    fig = px.line(test, x="timeline", y="chance", color="disease", markers=True, line_shape="hv", color_discrete_map = color_map) 
    # fig.update_traces(line_width=7)

    fig.update_yaxes(title_text="Survival rate")
    fig.update_xaxes(title_text="Time (months)")

    fig.update_xaxes(range=[-1, 60])
    return fig 



import numpy as np
import pandas as pd


def filter_data(df, th, at_least_good_cols=3, idx_cols=None):
    """Filter a given DataFrame by getting removing of the rows that have elements <= threshold.

    Args:
        df ([DataFrame]): The DataFrame from where we have to filter the data. This is has the genes as columns and samples as rows
        th ([float]): Threshold value of the unexpressed genes
        at_least_good_cols (int, optional): [The number of samples that fulfill the conditions]. Defaults to 3.

    Returns:
        [DataFrame]: DataFrame
    """
    if idx_cols is None:
        idx_cols = ["genes"]
        
    # eliminating the first column
    df_prcsd = df.drop(idx_cols, axis=1)
    # compute the selected genes
    selected_genes_idxs = df_prcsd[df_prcsd >= th].dropna(thresh=at_least_good_cols).index.values
    selected_genes = df_prcsd.iloc[selected_genes_idxs]
    # add the genes names back
    cols = [df.loc[selected_genes_idxs, idx_col] for idx_col in idx_cols]
    cols.append(selected_genes)
    selected_genes = pd.concat(cols, axis=1)

    # reset indexes
    selected_genes.reset_index(drop=True, inplace=True)
    return selected_genes


def select_genes(tpm_df, no_genes=3347, good_th=0.5, relative_selection=True, use_median=True):
    """
     It selects the most relative varied genes in the given DataFrame for the given number

    Args:
        tcga_tpm_df ([Dataframe]): The dataframe from where to select
        no_genes_selected (int, optional): [Genes to select]. Defaults to 3347.

    Returns:
        [type]: [description]
    """

    # remove all the genes w/ that have a lower expression value from `th` in `>10%` across the samples
    # dummy_df = pd.concat([pd.DataFrame(tcga_tpm_df["genes"]), pd.DataFrame(np.log2(tcga_tpm_df.iloc[:, 1:] + 1))], axis=1)
    # dummy_df = filter_data(dummy_df, th=np.log2(1.5), at_least_good_cols=dummy_df.shape[1] * 0.9, idx_cols=["genes"])

    dummy_df = np.log2(tpm_df.set_index("genes") + 1)
    at_least_good_cols = round(dummy_df.shape[1] * good_th)
    print(f"For th {good_th} ==> at least non-NAN values {at_least_good_cols}")

    # Do the filtering
    dummy_df = dummy_df[dummy_df >= np.log2(1.5)].dropna(thresh=at_least_good_cols).copy(deep=True)

    # acros samples
    if relative_selection:

        dmy = dummy_df.median(axis=1)
        if not use_median:
            dmy = dummy_df.mean(axis=1)

        print("The genes selected by the highest standard deviation/median ration.")
        dummy_df["ratio"] = dummy_df.std(axis=1) / dmy
    else:
        print("The genes selected by the highest standard deviation; approached used by Robertson et al.")
        dummy_df["ratio"] = dummy_df.std(axis=1)

    most_varied_genes = list(dummy_df.sort_values(by="ratio", ascending=False).iloc[:no_genes].index)
    print(f"####### Gene selection, num genes: {no_genes}. Prct included {no_genes/dummy_df.shape[0] * 100:.2f} % #######")

    return most_varied_genes


def remove_duplicates(df):
    """
       Removes the samples that are duplicates (i.e. resamples and are marked by having 01A#). We want to keep the ones with 01B
    Args:
        df (DataFrame): TCGA dataframe (columns - samples, rows - genes)

    Returns:
        list: The list of duplicates
    """
    # get the ones that are duplicated
    duplicates = ["-".join(col.split("-")[:-1]) for col in df if "01B" in col]
    # get with their pairs
    to_remove = []
    for col in df.columns:
        splits = "-".join(col.split("-")[:-1])
        if splits not in duplicates:
            continue

        if "01A" in col:
            to_remove.append(col)

    return to_remove


def create_map_cols(tcga_tpm_df):
    """
     Remove the -01B and -01A - this needs to be run only once

    Args:
        tcga_tpm_df ([DataFrame]): where to remove

    Returns:
        [Dict]: Dictionary of the old vs new col name
    """
    mapping_cols = {}
    mapping_cols["genes"] = "genes"
    for col in tcga_tpm_df.columns.values[1:]:
        mapping_cols[col] = "-".join(col.split("-")[:-1])
    return mapping_cols


def transp_df(df, samples_label="Samples"):
    """Transpose a given df and makes the original df's first row the header

    Args:
        df ([DataFrame]): Dataframe to transpose
        samples_label (str, optional): [description]. Defaults to "Samples".

    Returns:
        [DataFrame]: The transposed dataframe
    """
    df_t = df.transpose()
    df_t.columns = df.iloc[:, 0]
    df_t.drop(df_t.index[0], inplace=True)
    df_t.index.names = [samples_label]
    df_t.reset_index(inplace=True)
    df_t.rename_axis("Index", axis=1, inplace=True)
    return df_t


def prep_data(tcga_tpm_df, tcga_metadata_df, consensus_classifier, num_genes=3500, at_least_good=0.9, remap_cols=True):
    samples_to_remove = remove_duplicates(tcga_tpm_df)

    # drop from metadata, tpm
    tcga_tpm_df.drop(samples_to_remove, axis=1, inplace=True)

    tcga_metadata_df.drop(samples_to_remove, axis=1, inplace=True)

    ##### Apply pre-processing by Robertson et al. #####
    most_varied_genes_3k = select_genes(tcga_tpm_df, no_genes=num_genes, relative_selection=True, good_th=at_least_good)

    ##### The different genes #####
    selected_genes = set(most_varied_genes_3k)
    print(len(most_varied_genes_3k))

    ##### Sync DataFrames #####
    metadata_t = transp_df(tcga_metadata_df)
    samples_tpm = ["-".join(sample.split("-")[:-1]) for sample in metadata_t["Samples"].values]

    # remove the -01B and -01A - this needs to be run only once
    if remap_cols:
        mapping_cols = create_map_cols(tcga_tpm_df)
        working_tpm = tcga_tpm_df.copy(deep=True).rename(columns=mapping_cols)
    else:
        mapping_cols = list(tcga_tpm_df.columns)
        working_tpm = tcga_tpm_df.copy(deep=True)

    # These are duplicated samples with 01A and 01B
    common_samples = set(samples_tpm) - set(["TCGA-BL-A0C8", "TCGA-BL-A13J", "TCGA-FJ-A3Z9", "TCGA-BL-A13I"])

    # don't modify the original Dataframes
    working_metadata = metadata_t.copy(deep=True)

    working_metadata["Samples"] = samples_tpm
    working_metadata = working_metadata[working_metadata["Samples"].isin(common_samples)]
    # working_tpm.rename(columns=mapping_cols, inplace=True)
    working_tpm = working_tpm[["genes"] + list(common_samples)]

    # select only the common samples
    raw_metadata_t = working_metadata[working_metadata["Samples"].isin(common_samples)]

    # prepare data
    data_tpm = working_tpm.loc[tcga_tpm_df["genes"].isin(selected_genes)].copy(deep=True)
    data_tpm.set_index("genes", inplace=True)

    #
    consensus_classifier = consensus_classifier[consensus_classifier["Sample"].isin(common_samples)].set_index("Sample")
    raw_metadata_t.set_index("Samples", inplace=True)
    # raw_metadata_t.loc["2019_consensus_classifier"] = consensus_classifier["consensusClass_remapTCGA"]
    # raw_metadata_t.loc[consensus_classifier.index, "2019_consensus_classifier"] = consensus_classifier["consensusClass_remapTCGA"]

    # raw_metadata_t = pd.concat([raw_metadata_t, consensus_classifier[["consensusClass_remapTCGA", "TCGA_Robertson2017"]]], axis=1).rename(columns={
    #     "consensusClass_remapTCGA": "2019_consensus_classifier",
    #      "TCGA_Robertson2017":"TCGA408_classifier"}
    #      )

    raw_metadata_t["TCGA408_classifier"] = consensus_classifier["TCGA_Robertson2017"]
    raw_metadata_t["TCGA_2017_AM_remap"] = consensus_classifier["consensusClass_remapTCGA"]
    raw_metadata_t.reset_index(inplace=True)

    return data_tpm, working_tpm, raw_metadata_t, selected_genes, list(common_samples)



import os
import pickle

import pandas as pd

from .utilities import sankey_consensus_plot as sky
from .NetworkOutput import NetworkOutput


class ExperimentSet:

    def __init__(self, label, base_path, exp_path, mut_df, sel_sets=None, rel_path="", exp_type="PGCNA"):
        self.name = label
        self.path = os.path.join(base_path, exp_path)

        # read the master file and the experiments
        if exp_type == "PGCNA":
            raw_exps, self.master = ExperimentSet.read_experiments(self.path, sel_sets=sel_sets)
        else:
            raw_exps, self.master = ExperimentSet.read_experiments(self.path, sel_sets=sel_sets, exp_type="iNet")

        # check if everything is ok
        if len(raw_exps) != self.master.shape[0]:
            raise ImportError("Different length between master df and number of experiments")

        # unwrap experiments
        exps = {}
        for idx, exp_meta in self.master.iterrows():
            # just an extra check to see if things are ok
            if raw_exps[idx]["expName"] != exp_meta["expName"]:
                print("❗️ Check this experiment, there might be a problem in loading the data: {}".format(exp_meta["expName"]))

            # we need the base path for iNET only exps
            exp = NetworkOutput(raw_exps[idx]["graph"], exp_meta, mut_df, self.path, rel_path=rel_path, exp_type=exp_type, base_path=base_path)

            if "tf_list" in list(raw_exps[idx].keys()):
                exp.tf_list = raw_exps[idx]["tf_list"]

            key_val = exp.type
            exps[key_val] = exp
        self.exps = exps
        print("##### Experiment labels: ", exps.keys())

    @staticmethod
    def read_experiments(base_path, sel_sets=None, path="Stats/", exp_type="PGCNA"):
        """
        Each PCGNA run is stored in a master.csv. This function reads that .csv and returns a dataframe

        Args:
            path (_type_): _description_

        Returns:
            _type_: _description_
        """
        stats_path = base_path + path
        df = pd.read_csv(stats_path + "stats_master.tsv", sep="\t")
        # rename columns if exps are from iNet
        if exp_type == "iNet":
            remap_cols = {
                "exp_name": "expName",
                'modifier_type': "modifierType",
                "edges_pg": "edgesPG",
                "edges_sel": "edgesTF",
                "resolution_parameter": "resolution",
                "retain_f": "retainF",
                "genes_kept": "genesKept",
            }

            df.rename(columns=remap_cols, inplace=True)

        prcs_df = pd.DataFrame()
        exps = []
        for _, row in df.iterrows():
            # TODO the below code can be and needs to be improved!
            if sel_sets:
                toSkip = True
                for sel_set in sel_sets:
                    if row["genesKept"] == int(sel_set[0]) * 1000:  # 0 - num_genes, 1 - K
                        toSkip = False
                        break
                if toSkip:
                    continue

            picklePath = os.path.join(stats_path, row["expName"] + ".pickle")
            if os.path.exists(picklePath):
                with open(picklePath, "rb") as handle:
                    raw_exp = pickle.load(handle)

                    if exp_type == "iNet":
                        new_exp = {}
                        new_exp["expName"] = raw_exp["exp_name"]
                        new_exp["edgesPG"] = raw_exp["edges_pg"]
                        new_exp["edgesTF"] = raw_exp["edges_sel"]
                        new_exp["resolution"] = raw_exp["retain_f"]
                        new_exp["retainF"] = raw_exp["retain_f"]
                        new_exp["filesUsed"] = raw_exp["input_ge_file"]
                        new_exp["graph"] = raw_exp["graph"]
                        new_exp["genesKept"] = raw_exp["genes_kept"]
                        new_exp["tf_list"] = raw_exp["tf_list"]

                        exps.append(new_exp)
                    else:
                        exps.append(raw_exp)
            else:
                print("No file at {}".format(picklePath))

            prcs_df = pd.concat([prcs_df, row], axis=1)

        return exps, prcs_df.transpose().reset_index()

    def get_exp_labels(self):
        return list(self.exps.keys())

    def get_exps(self):
        return self.exps.values()

    def generate_modConsMut(self, modifiers=[]):
        # The problem with the parallelized code is that it doesn't take the reference to the next code
        #  processes = []

        # for exp in self.exps.values():
        #     process = mp.Process(target=exp.get_ModCon_Mut, args=(modifiers,))
        #     processes.append(process)
        #     process.start()

        # # wait for all processes to complete
        # for process in processes:
        #     process.join()
        # # report that all tasks are completed
        # print('Done', flush=True)
        for exp in self.exps.values():
            # exp.get_ModCon_Mut(modifiers)
            exp.get_ModCon()

    def generate_Mevs(self, no_genes=100):
        for exp in self.get_exps():
            if "beta2" in exp.name or "beta3" in exp.name:
                continue

            sort_col = "ModCon_{}".format(exp.type)
            exp.mevsMut, _ = exp.get_mevs(exp.tpm_df, exp.modCons, sort_col=sort_col, num_genes=100)

    def combine_edges(self):
        combined_edges = pd.DataFrame()
        for idx, exp in enumerate(self.get_exps()):
            col = "Weight_{}".format(exp.type)
            exp.mergeSourceTarget(col_name=col)

            if idx == 0:
                first_exp = self.get_exp_labels()[0]
                combined_edges = self.exps[first_exp].edges_df[["Source", "Target", col]].copy(deep=True)
            else:
                combined_edges = pd.concat([combined_edges, exp.edges_df[col]], axis=1)

        # populate all the source-target based on the index. This is needed as the experiments have different number of edges

        source, target = [], []
        for index in combined_edges.index:
            split = index.split("-")
            source.append(split[0])
            target.append(split[1])
        combined_edges["Source"] = source
        combined_edges["Target"] = target

        self.combined_edges = combined_edges
        return combined_edges

    def combine_nodes(self, mod_type="Leiden"):
        combined_nodes, comm_class = pd.DataFrame(), []

        if mod_type == 'Leiden':
            for idx, exp in enumerate(self.get_exps()):
                col = f"Leiden_{exp.type}"
                dmy_df = exp.nodes_df.rename(columns={"Modularity Class": col})
                dmy_df[col] = exp.type + "_" + dmy_df[col].astype(str)
                if idx == 0:
                    combined_nodes = dmy_df.copy(deep=True)
                else:
                    combined_nodes = pd.concat([combined_nodes, dmy_df[col].astype(str)], axis=1)

                comm_class.append(col)

                combined_nodes.fillna("NA", inplace=True)
        else:
            for idx, exp in enumerate(self.get_exps()):
                col = f"max_b_{exp.type}"
                dmy_df = exp.nodes_df.rename(columns={"max_b": col})
                dmy_df[col] = exp.type + "_" + dmy_df[col].astype(str)
                if idx == 0:
                    combined_nodes = dmy_df[["count", "ctrl_tf", col]].copy(deep=True)
                else:
                    combined_nodes = pd.concat([combined_nodes, dmy_df[col].astype(str)], axis=1)

                comm_class.append(col)

            # For pyarrow compatibility
            combined_nodes.fillna(pd.NA, inplace=True)

        self.combined_nodes = combined_nodes
        return combined_nodes, comm_class

    def combine_network_metric(self, stats):
        metrics = []
        for metric in stats:
            metric_df = pd.DataFrame()
            for key, val in self.exps.items():
                metric_df = pd.concat(
                    [
                        metric_df,
                        val.graph_stats[metric].rename("{}_{}".format(key, metric)),
                    ],
                    axis=1,
                )

            metrics.append(metric_df)

        self.networks_metrics = pd.concat(metrics, axis=1)
        return self.networks_metrics

    def export_to_gephi(self, label=None, save=True):
        if label:
            self.exps[label].export_to_gephi(save=save)
        else:
            for exp in self.exps.values():
                exp.export_to_gephi(save=save)

    def export_to_graphml(self):
        for exp in self.get_exps():
            filename = exp.name
            exp.graph.write_graphml("{}/PGCNA/{}/Graph_{}.graphml".format(exp.exps_path, filename, filename))

    def align_communities(self, source_exp, target_exp, mod_type):
        combined_nodes = self.combined_nodes
        # generate the mapping of experiments
        if mod_type == "Leiden":
            reoder_cols = ["Leiden_{}".format(source_exp), "Leiden_{}".format(target_exp)]
        else:
            reoder_cols = ["max_b_{}".format(source_exp), "max_b_{}".format(target_exp)]
            combined_nodes[reoder_cols] = combined_nodes[reoder_cols].fillna("NA")
        meta, _ = sky.main(
            combined_nodes,
            reorder_cols=reoder_cols,
            title="Community comparison between " + ", ".join(reoder_cols),
            retMeta=True,
        )

        # pre-process the data
        sourceTargetDf, labels = meta[0], meta[1]
        # below are the mapped labels of the groups used in Sankey (k) and the unique number given by the algorithm
        mapped_values = {k: v for k, v in enumerate(labels)}

        sourceTargetDf["SourceLabel"] = sourceTargetDf["sourceID"].map(mapped_values)
        sourceTargetDf["TargetLabel"] = sourceTargetDf["targetID"].map(mapped_values)

        # 3 Gives the number of paramaters that are taken in account for each experiment
        sourceTargetDf["sourceExpType"] = ["_".join(val.split("_")[:3]) for val in sourceTargetDf["SourceLabel"].values]
        sourceTargetDf["targetExpType"] = ["_".join(val.split("_")[:3]) for val in sourceTargetDf["TargetLabel"].values]

        # It is important to sort this by the number of changes ("count)" so that we get the remapping where the communities matched the most
        dmy = sourceTargetDf[sourceTargetDf["sourceExpType"] == source_exp].sort_values(by="count", ascending=False).copy(deep=True)
        remap_dict, checked_communities = {}, []
        for _, row in dmy.iterrows():
            if row["TargetLabel"] == "NA":
                continue

            source = int(row["SourceLabel"].split("_")[-1])
            target = int(row["TargetLabel"].split("_")[-1])

            if target in checked_communities:
                continue

            checked_communities.append(target)
            remap_dict[int(target)] = int(source)

        # Add the missing values. Otherwise the replace function won't work properly
        max_label = self.exps[target_exp].nodes_df["Modularity Class"].max()
        for label in range(1, max_label + 1):
            if label not in remap_dict.keys():
                remap_dict[label] = label

        node_df = self.exps[target_exp].nodes_df.copy(deep=True)

        col_name = "Leiden_r_{}".format(target_exp)
        node_df = node_df.replace({"Modularity Class": remap_dict}).rename(columns={"Modularity Class": col_name})
        tst_df = pd.concat([combined_nodes, node_df[col_name]], axis=1)
        tst_df[col_name] = tst_df[col_name].fillna("NA")
        tst_df[col_name] = "r_" + target_exp + "_" + tst_df[col_name].astype(str)

        return tst_df[col_name], sourceTargetDf

    def run_comb_exp(self, no_genes, egTF=50, comps=[], mod_type='Leiden'):
        # get the combinations of experiments
        if not len(comps):
            exp_labels = self.get_exp_labels()
            comps = [
                ("standard_{}K_{}TF".format(no_genes, egTF), label) for label in exp_labels if ("standard" not in label) and ("{}K".format(no_genes) in label)
            ]

        # combine the nodes
        if not hasattr(self, "combined_nodes"):
            _, _ = self.combine_nodes(mod_type=mod_type)

        # remap for all experimends
        remap_nodes_df = self.combined_nodes.copy(deep=True)
        meta_dict = {}
        for source_exp, target_exp in comps:
            remap_series, sourceTarget = self.align_communities(source_exp, target_exp, mod_type=mod_type)
            remap_nodes_df = pd.concat([remap_nodes_df, remap_series], axis=1)
            meta_dict["{}-{}".format(source_exp, target_exp)] = sourceTarget.rename(columns={"count": "changes_comm"})

        return remap_nodes_df, meta_dict

    def comb_leiden_scores(self):
        leiden_scores = pd.DataFrame()
        for exp in self.get_exps():
            exp.leiden_top3["Exp"] = exp.type
            exp.leiden_top3["TF"] = NetworkOutput.extract_tf_number(exp.type)
            exp.leiden_top3["Modifier"] = exp.type.split("_")[0]
            if "Leiden Rank" not in exp.leiden_top3.columns:
                exp.leiden_top3.reset_index(names="Leiden Rank", inplace=True)

            exp.leiden_top3["AvgModuleNum"] = exp.leiden_top3["ModuleNum"].mean()
            leiden_scores = pd.concat([leiden_scores, exp.leiden_top3], axis=0)

        # # rename beta3 to beta
        leiden_scores.loc[leiden_scores["Modifier"] == "beta2", "Modifier"] = "Penalty"
        leiden_scores.loc[leiden_scores["Modifier"] == "norm3", "Modifier"] = "Reward"
        leiden_scores.loc[leiden_scores["Modifier"] == "standard", "Modifier"] = "Standard"

        return leiden_scores

    def avg_leiden_scores(self):
        leiden_means, all_leiden = [], pd.DataFrame()
        for exp in self.get_exps():
            lg_df = pd.read_csv(exp.pgcna_path + "/LEIDENALG/moduleInfoBest.txt", sep="\t")

            # processing the data
            tf = exp.type.split("_")[-1].split("TF")[0]
            modifier = exp.type.split("_")[0]
            lg_df["Exp"] = exp.type
            lg_df["TF"] = tf
            lg_df["Modifier"] = modifier

            all_leiden = pd.concat([all_leiden, lg_df.drop(columns=["ModuleSizes"])], axis=0)
            leiden_means.append((exp.type, tf, modifier, lg_df["ModularityScore"].mean(), lg_df["Mod#"].mean()))

        stats_df = pd.DataFrame(leiden_means, columns=["Exp", "TF", "Modifier", "Avg_ModularityScore", "Avg_Mod#"])
        stats_df.loc[stats_df["Modifier"] == "beta2", "Modifier"] = "Penalty"
        stats_df.loc[stats_df["Modifier"] == "norm3", "Modifier"] = "Reward"
        stats_df.loc[stats_df["Modifier"] == "standard", "Modifier"] = "Standard"
        stats_df["Avg_Mod#"] = stats_df["Avg_Mod#"].round()

        all_leiden.loc[all_leiden["Modifier"] == "beta2", "Modifier"] = "Penalty"
        all_leiden.loc[all_leiden["Modifier"] == "norm3", "Modifier"] = "Reward"
        all_leiden.loc[all_leiden["Modifier"] == "standard", "Modifier"] = "Standard"
        return stats_df, all_leiden

    def comp_modConRank(self):
        for exp in self.get_exps():
            for modCon, value in exp.modCons.items():
                dmy = value.sort_values(by=["ModCon_{}".format(exp.type)], ascending=False).reset_index(names="Id").iloc[:100]
                dmy["Rank"] = dmy.index + 1
                dmy.set_index("Id", inplace=True)
                exp.nodes_df.loc[exp.nodes_df["Modularity Class"] == modCon, "ModCon_Rank"] = dmy["Rank"]
                exp.nodes_df["ModCon_Rank"] = exp.nodes_df["ModCon_Rank"].fillna(0)

    ##### Clustering methods #####
    def get_clusters(self, vu_output=None, tf=10, showFigs=False):
        comb_std = self.exps["standard_4K_{}TF".format(tf)].run_clusters(label="std_tf{}".format(tf), show_figs=showFigs)
        comb_norm3 = self.exps["norm3_4K_{}TF".format(tf)].run_clusters(label="norm3_tf{}".format(tf), show_figs=showFigs)
        comb_norm3.drop(columns=["PC_1", "PC_2"], inplace=True)
        comb_beta = self.exps["beta_4K_{}TF".format(tf)].run_clusters(label="beta_tf{}".format(tf), show_figs=showFigs)
        comb_beta.drop(columns=["PC_1", "PC_2"], inplace=True)

        dfs = [comb_std, comb_norm3, comb_beta]
        if vu_output is not None:
            dfs.append(vu_output)

        comb_all = pd.concat(dfs, axis=1).dropna()
        return comb_all